"""A module containing tests for data validation."""
import pytest
import iati.data
import iati.default
import iati.schemas
import iati.tests.utilities
import iati.validator


class ValidationTestBase(object):
    """A container for fixtures and other functionality useful among multiple groups of Validation Test."""

    @pytest.fixture
    def schema_basic(self):
        """An Activity Schema with no Codelists added."""
        return iati.default.activity_schema(None, False)

    @pytest.fixture(params=[
        iati.tests.utilities.load_as_string('valid_not_iati'),
        iati.tests.utilities.load_as_string('valid_iati'),
        iati.tests.utilities.load_as_string('valid_iati_invalid_code'),
        iati.tests.utilities.load_as_string('leading_whitespace_xml')
    ])
    def xml_str(self, request):
        """A valid XML string."""
        return request.param

    @pytest.fixture
    def xml_str_no_text_decl(self, xml_str):
        """A valid XML string with the text declaration removed."""
        return '\n'.join(xml_str.strip().split('\n')[1:])

    @pytest.fixture(params=iati.tests.utilities.generate_test_types([], True) + [iati.tests.utilities.load_as_string('invalid')])
    def not_xml(self, request):
        """A value that is not a valid XML string."""
        return request.param

    @pytest.fixture(params=['This is a string that is not XML.'])
    def str_not_xml(self, request):
        """A string that is not XML.

        Note:
            Does not use the utility function due to problems with Python 2.7.
        """
        return request.param

    @pytest.fixture
    def empty_str(self):
        """An empty string."""
        return ''

    @pytest.fixture(params=[
        iati.tests.utilities.load_as_dataset('valid_iati'),
        iati.tests.utilities.load_as_dataset('valid_iati_invalid_code')
    ])
    def iati_dataset(self, request):
        """A Dataset that is valid against the IATI Schema."""
        return request.param

    @pytest.fixture(params=[
        iati.tests.utilities.load_as_dataset('valid_not_iati')
    ])
    def not_iati_dataset(self, request):
        """A Dataset that is not valid against the IATI Schema."""
        return request.param

    @pytest.fixture(params=[
        iati.tests.utilities.load_as_dataset('invalid_iati_missing_required_element'),
        iati.tests.utilities.load_as_dataset('invalid_iati_missing_required_element_from_common')
    ])
    def not_iati_dataset_missing_required_el(self, request):
        """A Dataset that is not valid against the IATI Schema because it is missing a required element."""
        return request.param

    @pytest.fixture(params=iati.resources.get_test_data_paths_in_folder('ssot-activity-xml-pass'))
    def iati_dataset_valid_from_ssot(self, request):
        """A `should-pass` Dataset from the SSOT."""
        return iati.resources.load_as_dataset(request.param)

    @pytest.fixture(params=iati.resources.get_test_data_paths_in_folder('ssot-activity-xml-fail'))
    def iati_dataset_invalid_from_ssot(self, request):
        """A `should-fail` Dataset from the SSOT."""
        return iati.resources.load_as_dataset(request.param)

    @pytest.fixture
    def error_log_empty(self):
        """An empty error log.

        Note: MUST NOT have any errors added to it.

        """
        return iati.validator.ValidationErrorLog()


class TestValidationError(object):
    """A container for tests relating to ValidationErrors."""

    def test_validation_error_init_no_name(self):
        """Test that a ValidationError cannot be created with no name provided."""
        with pytest.raises(TypeError):
            iati.validator.ValidationError()  # pylint: disable=no-value-for-parameter

    @pytest.mark.parametrize("invalid_err_name", iati.tests.utilities.generate_test_types([], True))
    def test_validation_error_init_bad_name_type(self, invalid_err_name):
        """Test that a ValidationError cannot be created with a name that does not exist."""
        with pytest.raises(ValueError):
            iati.validator.ValidationError(invalid_err_name)

    def test_error_log_init_valid_name(self):
        """Test that a ValidationError can be created when provided a valid name."""
        err_name = 'err-code-not-on-codelist'
        err = iati.validator.ValidationError(err_name)
        err_detail = iati.validator.get_error_codes()[err_name]

        assert isinstance(err, iati.validator.ValidationError)
        assert err.name == err_name
        assert err.info == err_detail['info']
        assert err.help == err_detail['help']
        # attributes set with `setattr()`, so not picked up by linter
        assert err.base_exception == ValueError  # pylint: disable=no-member
        assert err.category == err_detail['category']  # pylint: disable=no-member
        assert err.description == err_detail['description']  # pylint: disable=no-member


class TestValidationErrorLog(ValidationTestBase):  # pylint: disable=too-many-public-methods
    """A container for tests relating to Validation Error Logs."""

    @pytest.fixture
    def err_name(self):
        """The name of an error."""
        return 'err-code-not-on-codelist'

    @pytest.fixture
    def err_type(self, err_name):
        """The type of an error."""
        return iati.validator.get_error_codes()[err_name]['base_exception']

    @pytest.fixture
    def error(self, err_name):
        """An error."""
        return iati.validator.ValidationError(err_name)

    @pytest.fixture
    def warning_name(self):
        """The name of a warning."""
        return 'warn-code-not-on-codelist'

    @pytest.fixture
    def warning_type(self, warning_name):
        """The type of an error."""
        return iati.validator.get_error_codes()[warning_name]['base_exception']

    @pytest.fixture
    def unused_exception_type(self):
        """An exception type that is not covered by the ValidationErrors."""
        return MemoryError

    @pytest.fixture
    def warning(self, warning_name):
        """A warning."""
        return iati.validator.ValidationError(warning_name)

    @pytest.fixture
    def error_log(self):
        """A basic error log that is initially empty."""
        return iati.validator.ValidationErrorLog()

    @pytest.fixture
    def error_log_with_error(self, error):
        """An error log with an error added."""
        error_log = iati.validator.ValidationErrorLog()
        error_log.add(error)

        return error_log

    @pytest.fixture
    def error_log_with_warning(self, warning):
        """An error log with a warning added."""
        error_log = iati.validator.ValidationErrorLog()
        error_log.add(warning)

        return error_log

    @pytest.fixture
    def error_log_mixed_contents(self, error, warning):
        """An error log with both an error and a warning added."""
        error_log = iati.validator.ValidationErrorLog()
        error_log.add(error)
        error_log.add(warning)

        return error_log

    def test_error_log_init(self, error_log, error_log_empty):
        """Test that a validator ErrorLog can be created and acts as a set."""
        assert isinstance(error_log, iati.validator.ValidationErrorLog)
        assert error_log == error_log_empty
        assert not error_log.contains_errors()
        assert not error_log.contains_warnings()

    def test_error_log_add_errors(self, error_log_with_error, err_name, warning_name, err_type):
        """Test that errors are identified as errors when added to the error log."""
        assert len(error_log_with_error) == 1
        assert error_log_with_error.contains_errors()
        assert not error_log_with_error.contains_warnings()
        assert error_log_with_error.contains_error_called(err_name)
        assert not error_log_with_error.contains_error_called(warning_name)
        assert error_log_with_error.contains_error_of_type(err_type)

    def test_error_log_add_warnings(self, error_log_with_warning, err_name, warning_name, warning_type):
        """Test that warnings are not identified as errors when added to the error log."""
        assert len(error_log_with_warning) == 1
        assert not error_log_with_warning.contains_errors()
        assert error_log_with_warning.contains_warnings()
        assert not error_log_with_warning.contains_error_called(err_name)
        assert error_log_with_warning.contains_error_called(warning_name)
        assert error_log_with_warning.contains_error_of_type(warning_type)

    # pylint: disable=too-many-arguments
    def test_error_log_add_mixed(self, error_log_mixed_contents, err_name, warning_name, err_type, warning_type, unused_exception_type):
        """Test that a mix of errors and warnings are identified as such when added to the error log."""
        assert len(error_log_mixed_contents) == 2
        assert error_log_mixed_contents.contains_errors()
        assert error_log_mixed_contents.contains_warnings()
        assert error_log_mixed_contents.contains_error_called(err_name)
        assert error_log_mixed_contents.contains_error_called(warning_name)
        assert error_log_mixed_contents.contains_error_of_type(err_type)
        assert error_log_mixed_contents.contains_error_of_type(warning_type)
        assert not error_log_mixed_contents.contains_error_of_type(unused_exception_type)

    @pytest.mark.parametrize("not_validation_error_obj", iati.tests.utilities.generate_test_types([], True))
    def test_error_log_add_incorrect_type(self, error_log, not_validation_error_obj):
        """Test that you may only add ValidationErrors to a ValidationErrorLog."""
        with pytest.raises(TypeError):
            error_log.add(not_validation_error_obj)

    @pytest.mark.parametrize("potential_validation_error_obj", iati.tests.utilities.generate_test_types([], True) + [iati.validator.ValidationError('err-code-not-on-codelist')])
    def test_error_log_set_index_incorrect_type(self, error_log_with_warning, potential_validation_error_obj):
        """Test that you may not add values to an error log via index assignment."""
        with pytest.raises(TypeError):
            error_log_with_warning[0] = potential_validation_error_obj

    def test_error_log_equality_single_error(self, error_log_with_error, error_log_with_warning):
        """Test equality between a pair of ValidationErrorLogs.

        Each error log contains the same number of errors. Each has different errors.
        """
        assert not error_log_with_error == error_log_with_warning
        assert not error_log_with_warning == error_log_with_error
        assert error_log_with_error != error_log_with_warning
        assert error_log_with_warning != error_log_with_error

    def test_error_log_equality_extended_log(self, error_log, error_log_with_error):
        """Test equality between a pair of ValidationErrorLogs.

        One error log is empty. The other has errors in.
        The empty log is extended with the values in the log that has values.
        """
        assert error_log != error_log_with_error
        assert error_log_with_error != error_log

        error_log.extend(error_log_with_error)

        assert error_log == error_log_with_error
        assert error_log_with_error == error_log

    def test_error_log_extend_from_list(self, error_log, error, warning):
        """Test extending an error log with values from a list.

        The list contains ValidationErrors.
        """
        to_extend = [error, warning]
        error_log.extend(to_extend)

        assert len(error_log) == 2
        assert error in error_log
        assert warning in error_log

    def test_error_log_extend_from_error_log(self, error_log, error_log_mixed_contents):
        """Test extending an error log with values from another error log."""
        error_log.extend(error_log_mixed_contents)

        assert len(error_log) == 2

    @pytest.mark.parametrize("iterable", iati.tests.utilities.generate_test_types(['bytearray', 'iter', 'list', 'mapping', 'memory', 'range', 'set', 'str', 'tuple', 'view']))
    def test_error_log_extend_from_iterable(self, error_log, error_log_empty, iterable):
        """Test extending an error log with a iterable.

        None of the iterables contain ValidationErrors.
        """
        error_log.extend(iterable)

        assert error_log == error_log_empty

    @pytest.mark.parametrize("non_iterable", iati.tests.utilities.generate_test_types(['bytearray', 'iter', 'list', 'mapping', 'memory', 'range', 'set', 'str', 'tuple', 'view'], True))
    def test_error_log_extend_from_non_iterable(self, error_log, error_log_empty, non_iterable):
        """Test extending an error log with a non-iterable."""
        with pytest.raises(TypeError):
            error_log.extend(non_iterable)

        assert error_log == error_log_empty


class TestValidationAuxiliaryData(object):
    """A container for tests relating to auxiliary validation data."""

    def test_error_code_names(self):
        """Check that the names of error codes are all in the correct format."""
        for err_code_name in iati.validator.get_error_codes().keys():
            assert err_code_name.split('-')[0] in ['err', 'warn']

    def test_error_code_attributes(self):
        """Check that error codes have the required attributes."""
        expected_attributes = [
            ('base_exception', type),
            ('category', str),
            ('description', str),
            ('info', str),
            ('help', str)
        ]
        for err_code in iati.validator.get_error_codes().values():
            code_attrs = err_code.keys()
            for (attr_name, attr_type) in expected_attributes:
                assert attr_name in code_attrs
                assert isinstance(err_code[attr_name], attr_type)


class ValidateCodelistsBase(ValidationTestBase):
    """A container for fixtures required for Codelist validation tests."""

    @pytest.fixture
    def schema_version(self):
        """Return an Activity Schema with the Version Codelist added."""
        schema = iati.default.activity_schema(None, False)
        codelist = iati.default.codelist('Version')

        schema.codelists.add(codelist)

        return schema

    @pytest.fixture
    def schema_org_type(self):
        """Return an Activity Schema with the OrganisationType Codelist added."""
        schema = iati.default.activity_schema(None, False)
        codelist = iati.default.codelist('OrganisationType')

        schema.codelists.add(codelist)

        return schema

    @pytest.fixture
    def schema_incomplete_codelist(self):
        """Return an Activity Schema with an incomplete Codelist added."""
        schema = iati.default.activity_schema(None, False)
        codelist = iati.default.codelist('Country')

        schema.codelists.add(codelist)

        return schema

    @pytest.fixture
    def schema_short_mapping_codelist(self):
        """Return an Activity Schema with a Codelist that has a short `path` in the mapping file."""
        schema = iati.default.activity_schema(None, False)
        codelist = iati.default.codelist('Language')

        schema.codelists.add(codelist)

        return schema

    @pytest.fixture
    def schema_sectors(self):
        """Return an Activity Schema with the DAC Sector Codelists and appropriate vocabulary added."""
        schema = iati.default.activity_schema(None, False)

        codelist_1 = iati.default.codelist('SectorVocabulary')
        codelist_2 = iati.default.codelist('Sector')
        codelist_3 = iati.default.codelist('SectorCategory')

        schema.codelists.add(codelist_1)
        schema.codelists.add(codelist_2)
        schema.codelists.add(codelist_3)

        return schema


class TestValidationTruthyIATI(ValidationTestBase):
    """A container for tests relating to truthy validation of IATI data."""

    def test_basic_validation_valid(self, schema_basic):
        """Perform a super simple data validation against a valid Dataset."""
        data = iati.tests.utilities.load_as_dataset('valid_iati')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_basic)
        assert iati.validator.is_valid(data, schema_basic)

    def test_basic_validation_should_pass_from_ssot(self, iati_dataset_valid_from_ssot, schema_basic):
        """Perform check to see whether a parameter is valid IATI XML.

        The parameter is valid IATI XML. It is sourced from the SSOT.
        """
        assert iati.validator.is_iati_xml(iati_dataset_valid_from_ssot, schema_basic)

    def test_basic_validation_invalid(self, schema_basic):
        """Perform a super simple data validation against an invalid Dataset."""
        data = iati.tests.utilities.load_as_dataset('valid_not_iati')

        assert iati.validator.is_xml(data.xml_str)
        assert not iati.validator.is_iati_xml(data, schema_basic)
        assert not iati.validator.is_valid(data, schema_basic)

    def test_basic_validation_should_fail_from_ssot(self, iati_dataset_invalid_from_ssot, schema_basic):
        """Perform check to see whether a parameter is valid IATI XML.

        The parameter is not valid IATI XML. It is sourced from the SSOT.
        """
        assert not iati.validator.is_iati_xml(iati_dataset_invalid_from_ssot, schema_basic)

    def test_basic_validation_invalid_missing_required_element(self, schema_basic):
        """Perform a super simple data validation against a Dataset that is invalid due to a missing required element."""
        data = iati.tests.utilities.load_as_dataset('invalid_iati_missing_required_element')

        assert iati.validator.is_xml(data.xml_str)
        assert not iati.validator.is_iati_xml(data, schema_basic)
        assert not iati.validator.is_valid(data, schema_basic)

    def test_basic_validation_invalid_missing_required_element_from_common(self, schema_basic):
        """Perform a super simple data validation against a Dataset that is invalid due to a missing required element that is defined in iati-common.xsd."""
        data = iati.tests.utilities.load_as_dataset('invalid_iati_missing_required_element_from_common')

        assert iati.validator.is_xml(data.xml_str)
        assert not iati.validator.is_iati_xml(data, schema_basic)
        assert not iati.validator.is_valid(data, schema_basic)


class TestValidateIsXML(ValidationTestBase):
    """A container for tests checking whether a value is valid XML."""

    def test_xml_check_valid_xml(self, xml_str):
        """Perform check to see whether a parameter is valid XML. The parameter is valid XML."""
        assert iati.validator.is_xml(xml_str)

    def test_xml_check_empty_string(self, empty_str):
        """Perform check to ensure an empty string is not valid XML."""
        assert not iati.validator.is_xml(empty_str)

    def test_xml_check_not_xml(self, not_xml):
        """Perform check to see whether a parameter is valid XML. The parameter is not valid XML."""
        assert not iati.validator.is_xml(not_xml)

    def test_xml_check_valid_xml_in_dataset(self, xml_str):
        """Perform check to see whether a Dataset is deemed valid XML."""
        data = iati.Dataset(xml_str)

        assert iati.validator.is_xml(data)

    def test_xml_check_valid_xml_detailed_output(self, xml_str, error_log_empty):
        """Perform check to see whether a parameter is valid XML. The parameter is valid XML.
        Obtain detailed error output.
        """
        result = iati.validator.validate_is_xml(xml_str)

        assert result == error_log_empty

    def test_xml_check_valid_xml_comments_after_detailed_output(self, xml_str, str_not_xml, error_log_empty):
        """Perform check to see string a parameter is valid XML.

        The string is valid XML.
        There is a comment added after the XML.
        Obtain detailed error output.
        """
        comment = '<!-- ' + str_not_xml + ' -->'
        xml_with_comments = xml_str + comment

        result = iati.validator.validate_is_xml(xml_with_comments)

        assert result == error_log_empty

    def test_xml_check_valid_xml_str_comments_before_no_text_decl_detailed_output(self, xml_str_no_text_decl, str_not_xml, error_log_empty):
        """Perform check to see whether a string is valid XML.

        The string is valid XML.
        There is a comment added before the XML. There is no XML text declaration.
        Obtain detailed error output.
        """
        comment = '<!-- ' + str_not_xml + ' -->'
        xml_prefixed_with_comment = comment + xml_str_no_text_decl

        result = iati.validator.validate_is_xml(xml_prefixed_with_comment)

        assert result == error_log_empty

    def test_xml_check_valid_xml_in_dataset_detailed_output(self, xml_str, error_log_empty):
        """Perform check to see whether a Dataset is valid XML.

        Obtain detailed error output.
        """
        data = iati.Dataset(xml_str)

        result = iati.validator.validate_is_xml(data)

        assert result == error_log_empty

    @pytest.mark.parametrize("not_str", iati.tests.utilities.generate_test_types(['str'], True))
    def test_xml_check_not_str_detailed_output(self, not_str):
        """Perform check to see whether a parameter is valid XML. The parameter is not valid XML.

        Obtain detailed error output.
        """
        result = iati.validator.validate_is_xml(not_str)

        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-not-string')

    def test_xml_check_not_xml_str_no_start_tag_detailed_output(self, str_not_xml):
        """Perform check to locate the XML Syntax Errors in a string.

        The string has no XML start tag.
        Obtain detailed error output.
        """
        result = iati.validator.validate_is_xml(str_not_xml)

        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-empty-document')

    def test_xml_check_not_xml_str_text_before_xml_detailed_output(self, str_not_xml, xml_str):
        """Perform check to locate the XML Syntax Errors in a string.

        The string has non-XML text before the XML starts.
        Obtain detailed error output.
        """
        not_xml = str_not_xml + xml_str

        result = iati.validator.validate_is_xml(not_xml)

        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-empty-document')

    def test_xml_check_not_xml_str_comments_before_detailed_output(self, xml_str, str_not_xml):
        """Perform check to locate the XML Syntax Errors in a string.

        There is a comment added before the XML. The XML contains a text declaration.
        Obtain detailed error output.
        """
        comment = '<!-- ' + str_not_xml + ' -->'
        not_xml = comment + xml_str

        result = iati.validator.validate_is_xml(not_xml)

        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-xml-text-decl-only-at-doc-start')

    def test_xml_check_not_xml_str_text_after_xml_detailed_output(self, xml_str, str_not_xml):
        """Perform check to locate the XML Syntax Errors in a string.

        The string has non-XML text before the XML starts.
        Obtain detailed error output.
        """
        not_xml = xml_str + str_not_xml

        result = iati.validator.validate_is_xml(not_xml)

        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-content-at-end')

    def test_xml_check_not_xml_str_xml_after_xml_detailed_output(self, xml_str):
        """Perform check to locate the XML Syntax Errors in a string.

        The string is two concatenated XML strings. Each contains a text declaration.
        Obtain detailed error output.
        """
        not_xml = xml_str + xml_str

        result = iati.validator.validate_is_xml(not_xml)

        assert len(result) == 2
        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-content-at-end')
        assert result.contains_error_called('err-not-xml-xml-text-decl-only-at-doc-start')

    def test_xml_check_not_xml_str_xml_after_xml_no_text_decl_detailed_output(self, xml_str_no_text_decl):
        """Perform check to locate the XML Syntax Errors in a string.

        The string is two concatenated XML strings. Each contains a text declaration.
        Obtain detailed error output.
        """
        not_xml = xml_str_no_text_decl + xml_str_no_text_decl

        result = iati.validator.validate_is_xml(not_xml)

        assert len(result) == 1
        assert result.contains_errors()
        assert result.contains_error_called('err-not-xml-content-at-end')


class TestIsValidIATIXML(ValidationTestBase):
    """A container for tests checking whether a value is valid IATI XML."""

    def test_iati_xml_check_valid_xml(self, iati_dataset, schema_basic, error_log_empty):
        """Perform check to see whether a parameter is valid IATI XML. The parameter is valid IATI XML."""
        result = iati.validator.validate_is_iati_xml(iati_dataset, schema_basic)

        assert result == error_log_empty

    def test_iati_xml_check_not_xml(self, not_iati_dataset, schema_basic):
        """Perform check to see whether a parameter is valid IATI XML.

        The parameter is not valid IATI XML.
        """
        result = iati.validator.validate_is_iati_xml(not_iati_dataset, schema_basic)

        assert result.contains_errors()

    def test_iati_xml_missing_required_element(self, not_iati_dataset_missing_required_el, schema_basic):  # pylint: disable=invalid-name
        """Perform check to see whether a parameter is valid IATI XML.

        The parameter is not valid IATI XML. It is missing a required element.
        """
        result = iati.validator.validate_is_iati_xml(not_iati_dataset_missing_required_el, schema_basic)

        assert result.contains_errors()
        assert result.contains_error_called('err-not-iati-xml-missing-required-element')

    def test_iati_xml_from_ssot_valid(self, schema_basic, error_log_empty):
        """Perform check to see whether valid XML from the SSOT can be loaded and validated."""
        data = iati.tests.utilities.load_as_dataset('ssot-activity-xml-pass/location/01-generic-location')

        result = iati.validator.validate_is_iati_xml(data, schema_basic)

        assert result == error_log_empty

    def test_iati_xml_from_ssot_invalid(self, schema_basic):
        """Perform check to see whether invalid XML from the SSOT can be loaded and validated."""
        data = iati.tests.utilities.load_as_dataset('ssot-activity-xml-fail/other-identifier/01-missing-attribute-type')

        result = iati.validator.validate_is_iati_xml(data, schema_basic)

        assert result.contains_errors()


class TestValidationCodelist(ValidateCodelistsBase):
    """A container for tests relating to validation of Codelists."""

    def test_basic_validation_codelist_valid(self, schema_version):
        """Perform data validation against valid IATI XML that has valid Codelist values."""
        data = iati.tests.utilities.load_as_dataset('valid_iati')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_version)
        assert iati.validator.is_valid(data, schema_version)

    def test_basic_validation_codelist_invalid(self, schema_version):
        """Perform data validation against valid IATI XML that has invalid Codelist values."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_invalid_code')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_version)
        assert not iati.validator.is_valid(data, schema_version)

    def test_basic_validation_codelist_valid_from_common(self, schema_org_type):
        """Perform data validation against valid IATI XML that has valid Codelist values. The attribute being tested is on an element defined in common.xsd."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_valid_code_from_common')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_org_type)
        assert iati.validator.is_valid(data, schema_org_type)

    def test_basic_validation_codelist_invalid_from_common(self, schema_org_type):
        """Perform data validation against valid IATI XML that has invalid Codelist values. The attribute being tested is on an element defined in common.xsd."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_invalid_code_from_common')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_org_type)
        assert not iati.validator.is_valid(data, schema_org_type)

    def test_basic_validation_codes_valid_multi_use_codelist(self, schema_org_type):
        """Perform data validation against valid IATI XML that has valid Codelist values. The attributes being tested are under different elements, but require the same Codelist."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_valid_codes_multiple_xpaths_for_codelist')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_org_type)
        assert iati.validator.is_valid(data, schema_org_type)

    @pytest.mark.parametrize("data", [iati.tests.utilities.load_as_dataset('valid_iati_invalid_codes_multiple_xpaths_for_codelist_first'), iati.tests.utilities.load_as_dataset('valid_iati_invalid_codes_multiple_xpaths_for_codelist_second')])
    def test_basic_validation_codes_invalid_multi_use_codelist(self, data, schema_org_type):
        """Perform data validation against valid IATI XML that has invalid Codelist values. The attributes being tested are under different elements, but require the same Codelist."""
        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_org_type)
        assert not iati.validator.is_valid(data, schema_org_type)

    def test_basic_validation_codelist_incomplete_present(self, schema_incomplete_codelist):
        """Perform data validation against valid IATI XML that has valid Codelist values. The attribute being tested refers to an incomplete Codelist. The value is on the list."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_incomplete_codelist_code_present')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_incomplete_codelist)
        assert iati.validator.is_valid(data, schema_incomplete_codelist)

    def test_basic_validation_codelist_incomplete_not_present(self, schema_incomplete_codelist):
        """Perform data validation against valid IATI XML that has valid Codelist values. The attribute being tested refers to an incomplete Codelist. The value is not on the list."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_incomplete_codelist_code_not_present')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_incomplete_codelist)
        assert iati.validator.is_valid(data, schema_incomplete_codelist)

    def test_basic_validation_short_mapping_xpath(self, schema_short_mapping_codelist):
        """Perform data validation against valid IATI XML. The attribute being tested refers to a Codelist with an abnormally short mapping file path. The data has no attributes mapped to by the Codelist."""
        data = iati.tests.utilities.load_as_dataset('valid_iati')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_short_mapping_codelist)
        assert iati.validator.is_valid(data, schema_short_mapping_codelist)


class TestValidationVocabularies(ValidateCodelistsBase):
    """A container for tests relating to validation of vocabularies and associated Codelists."""

    def test_validation_codelist_vocab_default_implicit(self, schema_sectors):
        """Perform data validation against valid IATI XML with a vocabulary that has been implicitly set."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_default_implicit')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_default_implicit_invalid_code(self, schema_sectors):
        """Perform data validation against valid IATI XML with a vocabulary that has been implicitly set. The code is invalid."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_default_implicit_invalid_code')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert not iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_default_explicit(self, schema_sectors):
        """Perform data validation against valid IATI XML with a vocabulary that has been explicitly set as the default value."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_default_explicit')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_non_default(self, schema_sectors):
        """Perform data validation against valid IATI XML with a vocabulary that has been explicitly set as a valid non-default value. The code is valid against this non-default vocabulary."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_non_default')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_multiple_same_valid(self, schema_sectors):
        """Perform data validation against valid IATI XML with an activity that uses multiple instances of the same element that uses vocabularies.

        The vocabulary used by each of these elements is the same.
        The codes are valid against the vocabularies.
        Percentages add up to 100.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_multiple_same_valid')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_multiple_different_valid(self, schema_sectors):
        """Perform data validation against valid IATI XML with an activity that uses multiple instances of the same element that uses vocabularies.

        The vocabulary used by each of these elements is different.
        The codes are valid against the vocabularies.
        Percentages add up to 100.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_multiple_different_valid')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_multiple_same_invalid_code(self, schema_sectors):
        """Perform data validation against valid IATI XML with an activity that uses multiple instances of the same element that uses vocabularies.

        The vocabulary used by each of these elements is the same.
        The codes are valid against the vocabularies.
        Percentages add up to 100.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_multiple_same_invalid_code')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert not iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_multiple_different_invalid_code(self, schema_sectors):
        """Perform data validation against valid IATI XML with an activity that uses multiple instances of the same element that uses vocabularies.

        The vocabulary used by each of these elements is different.
        The codes are valid against the vocabularies.
        Percentages add up to 100.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_multiple_different_invalid_code')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert not iati.validator.is_valid(data, schema_sectors)

    def test_validation_codelist_vocab_user_defined(self, schema_sectors):
        """Perform data validation against valid IATI XML with a user-defined vocabulary. No URI is defined, so the code cannot be checked."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_user_defined')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    @pytest.mark.skip(reason="Not yet implemented - need to be able to load Codelists from URLs")
    def test_validation_codelist_vocab_user_defined_with_uri_readable(self, schema_sectors):
        """Perform data validation against valid IATI XML with a user-defined vocabulary. A URI is defined, and points to a machine-readable codelist. As such, the code can be checked. The @code is valid."""
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_user_defined_with_uri_readable')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)

    @pytest.mark.skip(reason="Not yet implemented - need to be able to load Codelists from URLs")
    def test_validation_codelist_vocab_user_defined_with_uri_readable_bad_code(self, schema_sectors):
        """Perform data validation against valid IATI XML with a user-defined vocabulary. A URI is defined, and points to a machine-readable codelist. As such, the code can be checked. The @code is not in the list.

        Todo:
            Check that this is a legitimate check to be performed, given the contents and guidance given in the Standard.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_user_defined_with_uri_readable_bad_code')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert not iati.validator.is_valid(data, schema_sectors)

    @pytest.mark.skip(reason="Not yet implemented - need to be able to load Codelists from URLs")
    def test_validation_codelist_vocab_user_defined_with_uri_unreadable(self, schema_sectors):
        """Perform data validation against valid IATI XML with a user-defined vocabulary. A URI is defined, and points to a non-machine-readable codelist. As such, the @code cannot be checked. The @code is valid.

        Todo:
            Remove xfail and work on functionality to fully fetch and parse user-defined codelists after higher priority functionality is finished.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_vocab_user_defined_with_uri_unreadable')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_sectors)
        assert iati.validator.is_valid(data, schema_sectors)


class TestValidateRulesets(object):
    """A container for tests relating to validation of Rulesets."""

    def test_basic_validation_ruleset_valid(self, schema_ruleset):
        """Perform data validation against valid IATI XML that is valid against the Standard Ruleset."""
        data = iati.tests.utilities.load_as_dataset('valid_std_ruleset')

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_ruleset)
        assert iati.validator.is_valid(data, schema_ruleset)

    @pytest.mark.parametrize("invalid_xml_file", [
        'ruleset-std/invalid_std_ruleset_bad_date_order',
        'ruleset-std/invalid_std_ruleset_bad_identifier',
        'ruleset-std/invalid_std_ruleset_does_not_sum_100',
        'ruleset-std/invalid_std_ruleset_missing_sector_element'
    ])
    def test_basic_validation_ruleset_invalid(self, schema_ruleset, invalid_xml_file):
        """Perform data validation against valid IATI XML that does not conform to the Standard Ruleset."""
        data = iati.tests.utilities.load_as_dataset(invalid_xml_file)

        assert iati.validator.is_xml(data.xml_str)
        assert iati.validator.is_iati_xml(data, schema_ruleset)
        assert not iati.validator.is_valid(data, schema_ruleset)

    def test_one_ruleset_error_added_for_multiple_rule_errors(self, schema_ruleset):
        """Check that a Dataset containing multiple Rule errors produces an error log containing only one Ruleset error."""
        data_with_multiple_rule_errors = iati.tests.utilities.load_as_dataset('ruleset-std/invalid_std_ruleset_multiple_rule_errors')
        result = iati.validator.full_validation(data_with_multiple_rule_errors, schema_ruleset)

        assert len(result.get_errors_or_warnings_by_category('rule')) > 1
        assert len(result.get_errors_or_warnings_by_name('err-ruleset-conformance-fail')) == 1

    def test_multiple_ruleset_error_added_for_multiple_rulesets(self):
        """Check that a Schema containing multiple Rulesets produces an error log containing multiple Ruleset errors when each errors."""
        data_with_multiple_rule_errors = iati.tests.utilities.load_as_dataset('ruleset-std/invalid_std_ruleset_multiple_rule_errors')
        ruleset_1 = iati.default.ruleset()
        ruleset_2 = iati.default.ruleset()
        schema = iati.default.activity_schema(None, False)
        schema.rulesets.add(ruleset_1)
        schema.rulesets.add(ruleset_2)
        result = iati.validator.full_validation(data_with_multiple_rule_errors, schema)

        assert len(result.get_errors_or_warnings_by_category('rule')) > 1
        assert len(result.get_errors_or_warnings_by_name('err-ruleset-conformance-fail')) == 2

    def test_no_ruleset_errors_added_for_rule_warnings(self, schema_ruleset):
        """Check that a Dataset containing only Rule warnings does not result in a Ruleset error being added."""
        data_with_rule_warnings_only = iati.tests.utilities.load_as_dataset('valid_std_ruleset')
        result = iati.validator.full_validation(data_with_rule_warnings_only, schema_ruleset)

        assert len(result.get_warnings()) >= 1
        assert len(result.get_errors_or_warnings_by_category('rule')) >= 1
        assert result.get_errors_or_warnings_by_name('err-ruleset-conformance-fail') == []


class TestValidatorFullValidation(ValidateCodelistsBase):
    """A container for tests relating to detailed error output from validation."""

    def test_full_validation_codelist_valid_detailed_output(self, schema_version):
        """Perform data validation against valid IATI XML that has valid Codelist values.  Obtain detailed error output."""
        data = iati.tests.utilities.load_as_dataset('valid_iati')

        assert iati.validator.full_validation(data, schema_version) == iati.validator.ValidationErrorLog()

    def test_full_validation_codelist_invalid_detailed_output(self, schema_version):
        """Perform data validation against valid IATI XML that has invalid Codelist values."""
        xml_str = iati.tests.utilities.load_as_string('valid_iati_invalid_code')
        data = iati.Dataset(xml_str)

        result = iati.validator.full_validation(data, schema_version)[0]

        assert isinstance(result, iati.validator.ValidationError)
        assert result.name == 'err-code-not-on-codelist'
        assert result.status == 'error'
        assert result.line_number == 3
        assert result.context == '\n'.join(xml_str.split('\n')[1:4])
        assert 'Version' in result.info
        assert 'Version' in result.help

    def test_full_validation_codelist_incomplete_present_detailed_output(self, schema_incomplete_codelist, error_log_empty):
        """Perform data validation against valid IATI XML that has valid Codelist values. The attribute being tested refers to an incomplete Codelist. The value is on the list.
        Obtain detailed error output.
        """
        data = iati.tests.utilities.load_as_dataset('valid_iati_incomplete_codelist_code_present')

        result = iati.validator.full_validation(data, schema_incomplete_codelist)

        assert result == error_log_empty

    def test_full_validation_codelist_incomplete_not_present_detailed_output(self, schema_incomplete_codelist):
        """Perform data validation against valid IATI XML that has valid Codelist values. The attribute being tested refers to an incomplete Codelist. The value is not on the list.
        Obtain detailed error output.
        """
        xml_str = iati.tests.utilities.load_as_string('valid_iati_incomplete_codelist_code_not_present')
        data = iati.Dataset(xml_str)

        result = iati.validator.full_validation(data, schema_incomplete_codelist)[0]

        assert result.name == 'warn-code-not-on-codelist'
        assert result.line_number == 18
        assert result.context == '\n'.join(xml_str.split('\n')[16:19])
        assert result.status == 'warning'
        assert 'Country' in result.info
        assert 'Country' in result.help

    def test_full_validation_not_xml_detailed_output(self, schema_basic):
        """Perform full validation against a string that is not XML."""
        not_xml = 'This is not XML.'

        result = iati.validator.full_validation(not_xml, schema_basic)

        assert len(result) == 1
        assert result.contains_error_called('err-not-xml-empty-document')



"""A module containing tests for the library representation of Rulesets.

Todo:
    Remove references to `case`.
    Check all tests are still necessary/relevant.
    Condense tests for individual Rules.
    Try to standardise Rule arrays for extraction.

"""
# pylint: disable=protected-access,too-many-lines
from copy import deepcopy
import pytest
import iati.default
import iati.rulesets
import iati.resources
import iati.tests.utilities


class TestRuleset(object):
    """A container for tests relating to Rulesets."""

    def test_ruleset_init_no_parameters(self):
        """Check that an empty Ruleset can be created when no parameters are given."""
        ruleset = iati.Ruleset()

        assert isinstance(ruleset, iati.Ruleset)
        assert isinstance(ruleset.rules, set)
        assert ruleset.rules == set()

    @pytest.mark.parametrize("ruleset_str", [
        '{"CONTEXT": {"atleast_one": {"cases": []}}}',  # JSON string that has no Rules
        ' ',  # whitespace only
        '',  # empty string
        None  # none
    ])
    def test_ruleset_init_ruleset_valid_no_rules(self, ruleset_str):
        """Check that a Ruleset is created when given a JSON Ruleset in string format even if it contains no Rules."""
        ruleset = iati.Ruleset(ruleset_str)

        assert isinstance(ruleset, iati.Ruleset)
        assert isinstance(ruleset.rules, set)
        assert ruleset.rules == set()

    @pytest.mark.parametrize("not_a_ruleset", iati.tests.utilities.generate_test_types(['str', 'bytearray', 'none'], True))
    def test_ruleset_init_ruleset_str_not_str(self, not_a_ruleset):
        """Check that a Ruleset cannot be created when given at least one Rule in a non-string format."""
        with pytest.raises(ValueError):
            iati.Ruleset(not_a_ruleset)

    @pytest.mark.parametrize("byte_array", iati.tests.utilities.generate_test_types(['bytearray']))
    def test_ruleset_init_ruleset_str_bytearray(self, byte_array):
        """Check that a Ruleset cannot be created when given at least one Rule in a bytearray format."""
        with pytest.raises(ValueError):
            iati.Ruleset(byte_array)

    def test_ruleset_init_ruleset_str_invalid(self):
        """Check that a Ruleset cannot be created when given a string that is not a Ruleset."""
        not_a_ruleset_str = 'This is not a ruleset: It is a cat. Meow, meow.'

        with pytest.raises(ValueError):
            iati.Ruleset(not_a_ruleset_str)

    def test_ruleset_init_ruleset_1_rule(self):
        """Check that a Ruleset can be created when given a JSON Ruleset in string format with one Rule."""
        ruleset_str = '{"CONTEXT": {"atleast_one": {"cases": [{"paths": ["test_path"]}]}}}'

        ruleset = iati.Ruleset(ruleset_str)

        assert isinstance(ruleset, iati.Ruleset)
        assert isinstance(ruleset.rules, set)
        assert len(ruleset.rules) == 1
        assert isinstance(list(ruleset.rules)[0], iati.Rule)
        assert isinstance(list(ruleset.rules)[0], iati.RuleAtLeastOne)

    def test_ruleset_init_ruleset_1_rule_invalid_type(self):
        """Check that a Ruleset raises a KeyError when given a JSON Ruleset in string format with an invalid rule_type key."""
        ruleset_str = '{"CONTEXT": {"invalid_rule_type": {"cases": [{"paths": ["test_path"]}]}}}'

        with pytest.raises(ValueError):
            iati.Ruleset(ruleset_str)

    def test_ruleset_init_ruleset_2_rules_single_case(self):
        """Check that a Ruleset can be created when given a JSON Ruleset in string format with two Rules under a single case."""
        ruleset_str = '{"CONTEXT": {"atleast_one": {"cases": [{"paths": ["test_path_1"]}, {"paths": ["test_path_2"]}]}}}'

        ruleset = iati.Ruleset(ruleset_str)

        assert isinstance(ruleset, iati.Ruleset)
        assert isinstance(ruleset.rules, set)
        assert len(ruleset.rules) == 2
        for rule in ruleset.rules:
            assert isinstance(rule, iati.Rule)
            assert isinstance(rule, iati.RuleAtLeastOne)

    def test_ruleset_init_ruleset_multiple_cases(self):
        """Check that a Ruleset can be created when given a JSON Ruleset in string format with two Rules of different types, each under the same context."""
        ruleset_str = '{"CONTEXT": {"atleast_one": {"cases": [{"paths": ["test_path_1"]}]}, "no_more_than_one": {"cases": [{"paths": ["test_path_2"]}]}}}'

        ruleset = iati.Ruleset(ruleset_str)

        assert isinstance(ruleset, iati.Ruleset)
        assert isinstance(ruleset.rules, set)
        assert len(ruleset.rules) == 2
        for rule in ruleset.rules:
            assert isinstance(rule, iati.Rule)
        assert len([rule for rule in ruleset.rules if isinstance(rule, iati.RuleAtLeastOne)]) == 1
        assert len([rule for rule in ruleset.rules if isinstance(rule, iati.RuleNoMoreThanOne)]) == 1

    def test_ruleset_init_ruleset_duplicate_types(self):
        """Check that a Ruleset raises a ValueError when given a JSON Ruleset in string format with two Rules of the same type, each under the same context."""
        ruleset_str = '{"CONTEXT": {"atleast_one": {"cases": [{"paths": ["test_path_1"]}]}, "atleast_one": {"cases": [{"paths": ["test_path_2"]}]}}}'

        with pytest.raises(ValueError):
            iati.Ruleset(ruleset_str)

    def test_ruleset_init_ruleset_duplicate_contexts(self):
        """Check that a Ruleset raises a ValueError when given a JSON Ruleset in string format with two Rules of the same type, each under the same context."""
        ruleset_str = '{"DUPLICATE_CONTEXT": {"atleast_one": {"cases": [{"paths": ["test_path_1"]}]}}, "DUPLICATE_CONTEXT": {"no_more_than_one": {"cases": [{"paths": ["test_path_2"]}]}}}'

        with pytest.raises(ValueError):
            iati.Ruleset(ruleset_str)

    def test_ruleset_init_ruleset_multiple_contexts(self):
        """Check that a Ruleset can be created when given a JSON Ruleset in string format with two Rules of the same type, each under a different context."""
        ruleset_str = '{"CONTEXT_1": {"atleast_one": {"cases": [{"paths": ["test_path_1"]}]}}, "CONTEXT_2": {"atleast_one": {"cases": [{"paths": ["test_path_2"]}]}}}'

        ruleset = iati.Ruleset(ruleset_str)

        assert isinstance(ruleset, iati.Ruleset)
        assert isinstance(ruleset.rules, set)
        assert len(ruleset.rules) == 2
        for rule in ruleset.rules:
            assert isinstance(rule, iati.Rule)
            assert isinstance(rule, iati.RuleAtLeastOne)

    def test_ruleset_is_valid_for_valid_dataset(self):
        """Check that a Dataset can be validated against the Standard Ruleset."""
        ruleset = iati.tests.utilities.RULESET_FOR_TESTING
        valid_dataset = iati.tests.utilities.load_as_dataset('valid_std_ruleset')
        assert ruleset.is_valid_for(valid_dataset)

    @pytest.mark.parametrize("invalid_dataset", [
        iati.tests.utilities.load_as_dataset('ruleset-std/invalid_std_ruleset_bad_date_order'),
        iati.tests.utilities.load_as_dataset('ruleset-std/invalid_std_ruleset_bad_identifier'),
        iati.tests.utilities.load_as_dataset('ruleset-std/invalid_std_ruleset_does_not_sum_100'),
        iati.tests.utilities.load_as_dataset('ruleset-std/invalid_std_ruleset_missing_sector_element')
    ])
    def test_ruleset_is_invalid_for_invalid_dataset(self, invalid_dataset):
        """Check that a Dataset can be invalidated against the Standard Ruleset."""
        ruleset = iati.tests.utilities.RULESET_FOR_TESTING
        assert not ruleset.is_valid_for(invalid_dataset)

    @pytest.mark.parametrize("dataset_name, rule_type, case", [
        ('ruleset/invalid_format_dateorder', 'date_order', {'less': 'element1', 'more': 'element2'}),
        ('ruleset/invalid_startswith', 'startswith', {'start': 'duplicateprefix', 'paths': ['element12']}),
        ('ruleset/invalid_sum', 'sum', {'paths': ['element42'], 'sum': 50}),
    ])
    def test_ruleset_is_invalid_for_valueerror(self, dataset_name, rule_type, case):
        """Check that `ValueError`s are correctly handled when checking a Ruleset.

        Rulesets should absorb them and return `False` rather than passing them on to the caller.

        """
        invalid_dataset = iati.tests.utilities.load_as_dataset(dataset_name)
        rule_constructor = iati.rulesets.constructor_for_rule_type(rule_type)
        rule = rule_constructor('//root_element', case)
        ruleset = iati.Ruleset('')
        ruleset.rules.add(rule)

        assert not ruleset.is_valid_for(invalid_dataset)


class TestRule(object):
    """A container for tests relating to Rules."""

    def test_rule_class_cannot_be_instantiated_directly_without_name(self):
        """Check that Rule itself cannot be directly instantiated."""
        context = 'an xpath'
        case = {'paths': ['path_1', 'path_2']}

        with pytest.raises(AttributeError):
            iati.Rule(context, case)

    def test_rule_class_cannot_be_instantiated_directly_with_name(self):
        """Check that Rule itself cannot be directly instantiated with a Rule name."""
        name = 'atleast_one'
        context = 'an xpath'
        case = {'paths': ['path_1', 'path_2']}

        with pytest.raises(TypeError):
            iati.Rule(name, context, case)  # pylint: disable=too-many-function-args


class TestRuleSubclasses(object):
    """A container for tests relating to all Rule subclasses."""

    @pytest.fixture(params=[
        iati.rulesets.constructor_for_rule_type(rule_type) for rule_type in iati.rulesets._VALID_RULE_TYPES
    ])
    def rule_constructor(self, request):
        """Return constructor for the current type of Rule.

        Note:
            Differs from fixture of same name in RuleSubclassTestBase as specifies rule_type by _VALID_RULE_TYPES rather than via specific TestRuleSubclass.

        """
        return request.param

    def test_rule_init_no_parameters(self, rule_constructor):
        """Check that a Rule cannot be created when no parameters are given."""
        with pytest.raises(TypeError):
            rule_constructor()

    @pytest.mark.parametrize("xpath_base", iati.tests.utilities.generate_test_types(['str'], True))
    def test_rule_init_invalid_xpath_base(self, rule_constructor, xpath_base):
        """Check that a Rule cannot be created when xpath_base is not a string.

        Todo:
            Ensure the case is valid - an empty dictionary isn't.
        """
        case = dict()

        with pytest.raises(TypeError):
            rule_constructor(xpath_base, case)

    @pytest.mark.parametrize("case", iati.tests.utilities.generate_test_types(['mapping'], True))
    def test_rule_init_invalid_case(self, rule_constructor, case):
        """Check that a Rule cannot be created when case is not a dictionary."""
        context = 'an xpath'

        with pytest.raises(ValueError):
            rule_constructor(context, case)

    def test_rule_init_invalid_case_property(self, rule_constructor):
        """Check that a Rule cannot be created when a case has a property that is not permitted."""
        context = 'an xpath'
        case = {'thisis_an_invalidkey': ['this_is_a_value']}

        with pytest.raises(ValueError):
            rule_constructor(context, case)


class RuleSubclassTestBase(object):  # pylint: disable=too-many-public-methods
    """A base class for Rule subclass tests."""

    @pytest.fixture
    def valid_single_context(self):
        """Return valid context with a single match."""
        return '//root_element'

    @pytest.fixture
    def valid_multiple_context(self):
        """Return a valid context with multiple matches."""
        return '//nest'

    @pytest.fixture
    def non_existent_context(self):
        """Return an XPath for a context that does not exist."""
        return '//non-existent-context'

    @pytest.fixture(params=[
        'count(condition)>0',
        'condition'
    ])
    def condition_is_true_valid(self, validating_case, request):
        """Return a case with optional condition that evaluates to `True` for a valid dataset."""
        condition_validating_case = deepcopy(validating_case)
        condition_validating_case['condition'] = request.param
        return condition_validating_case

    @pytest.fixture(params=[
        'count(condition)>0',
        'condition'
    ])
    def condition_is_true_invalid(self, invalidating_case, request):
        """Return a case with optional condition that evaluates to `True` for an invalid dataset."""
        condition_invalidating_case = deepcopy(invalidating_case)
        condition_invalidating_case['condition'] = request.param
        return condition_invalidating_case

    @pytest.fixture(params=[
        'count(condition)<1',
        'nocondition'
    ])
    def condition_is_false_valid(self, validating_case, request):
        """Return a case with an optional condition that evaluates to False for a valid dataset."""
        condition_validating_case = deepcopy(validating_case)
        condition_validating_case['condition'] = request.param
        return condition_validating_case

    @pytest.fixture
    def rule_instantiating(self, rule_constructor, instantiating_case, valid_single_context):
        """Rule subclass that instantiates but is not used for validation testing."""
        return rule_constructor(valid_single_context, instantiating_case)

    @pytest.fixture
    def rule_valid(self, rule_constructor, validating_case, valid_single_context):
        """Rule for checking the `is_valid_for` function against a relevant valid Dataset."""
        return rule_constructor(valid_single_context, validating_case)

    @pytest.fixture
    def rule_invalid(self, rule_constructor, invalidating_case, valid_single_context):
        """Rule for checking the `is_valid_for` function against a relevant invalid Dataset."""
        return rule_constructor(valid_single_context, invalidating_case)

    @pytest.fixture
    def rule_constructor(self, rule_type):
        """Return current Rule type.

        Note:
            Differs from fixture of same name in TestRuleSubclass as specifies rule_type via specific TestRuleSubclass fixture.

        """
        return iati.rulesets.constructor_for_rule_type(rule_type)

    @pytest.fixture
    def valid_condition_rule(self, rule_constructor, valid_single_context, condition_is_true_valid):
        """Return a Rule with a `condition`."""
        return rule_constructor(valid_single_context, condition_is_true_valid)

    @pytest.fixture
    def invalid_condition_rule(self, rule_constructor, valid_single_context, condition_is_true_invalid):
        """Return a Rule with a `condition`."""
        return rule_constructor(valid_single_context, condition_is_true_invalid)

    def test_rule_init_valid_parameter_types(self, rule_instantiating):
        """Check that Rule subclasses can be instantiated with valid parameter types."""
        assert isinstance(rule_instantiating, iati.Rule)

    def test_rule_init_raises_error_on_empty_context(self, rule_constructor, instantiating_case):
        """Check that a Rule cannot be instantiated when the `context` is an empty string."""
        invalid_context = ''
        with pytest.raises(ValueError):
            rule_constructor(invalid_context, instantiating_case)

    def test_rule_attributes_from_case(self, rule_instantiating):
        """Check that a Rule subclass has mandatory case attributes set."""
        required_attributes = rule_instantiating._case_attributes(rule_instantiating._ruleset_schema_section())
        for attrib in required_attributes:
            # Ensure that the attribute exists - if not, an AttributeError will be raised
            getattr(rule_instantiating, attrib)

    def test_optional_rule_attributes_from_case(self, rule_constructor, valid_single_context, condition_is_true_valid):
        """Check that a Rule subclass has optional case attributes set."""
        rule = rule_constructor(valid_single_context, condition_is_true_valid)
        optional_attributes = rule._case_attributes(rule._ruleset_schema_section(), False)
        for attrib in optional_attributes:
            # Ensure that the attribute exists - if not, an AttributeError will be raised
            getattr(rule, attrib)

    def test_rule_name(self, rule_instantiating, rule_type):
        """Check that a Rule subclass has the expected name."""
        assert rule_instantiating.name == rule_type

    def test_rule_string_output_general(self, rule_instantiating):
        """Check that the string format of the Rule has been customised and variables formatted."""
        assert 'iati.rulesets' not in str(rule_instantiating)
        assert ' object at ' not in str(rule_instantiating)
        assert 'self' not in str(rule_instantiating)
        assert '{0}' not in str(rule_instantiating)
        assert 'This is a Rule' not in str(rule_instantiating)

    @pytest.mark.parametrize("context", iati.tests.utilities.generate_test_types(['str'], True))
    def test_rule_init_invalid_context(self, rule_constructor, context, instantiating_case):
        """Check that a Rule subclass cannot be created when context is not a string."""
        with pytest.raises(TypeError):
            rule_constructor(context, instantiating_case)

    def test_rule_invalid_case(self, rule_constructor, uninstantiating_case):
        """Check that a Rule cannot be instantiated when the case is not permitted."""
        context = 'an xpath'
        with pytest.raises(ValueError):
            rule_constructor(context, uninstantiating_case)

    def test_is_valid_for(self, valid_dataset, rule_valid):
        """Check that a given Rule returns the expected result when given Dataset."""
        assert rule_valid.is_valid_for(valid_dataset)

    def test_is_invalid_for(self, invalid_dataset, rule_invalid):
        """Check that a given Rule returns the expected result when given a Dataset."""
        assert not rule_invalid.is_valid_for(invalid_dataset)

    @pytest.mark.parametrize("junk_data", iati.tests.utilities.generate_test_types([], True))
    def test_is_valid_for_raises_error_on_non_permitted_argument(self, rule_instantiating, junk_data):
        """Check that a given Rule returns expected error when passed an argument that is not a Dataset."""
        with pytest.raises(TypeError):
            rule_instantiating.is_valid_for(junk_data)

    def test_is_valid_for_raises_error_when_passed_an_etree(self, rule_instantiating):
        """Check that an error is raised if an etree is given as an argument instead of a Dataset.

        Todo:
            Use more generic Dataset.

        """
        with pytest.raises(TypeError):
            rule_instantiating.is_valid_for(iati.resources.load_as_tree(iati.resources.get_test_data_path('ruleset/valid_atleastone')))

    def test_multiple_valid_context_matches_is_valid_for(self, valid_multiple_context, valid_nest_case, rule_constructor, valid_dataset):
        """Check Rule returns expected result when checking multiple contexts."""
        rule = rule_constructor(valid_multiple_context, valid_nest_case)
        assert rule.is_valid_for(valid_dataset)

    def test_multiple_valid_context_matches_is_invalid_for(self, valid_multiple_context, invalid_nest_case, rule_constructor, invalid_dataset):
        """Check Rule returns expected result when checking multiple contexts."""
        rule = rule_constructor(valid_multiple_context, invalid_nest_case)
        assert not rule.is_valid_for(invalid_dataset)

    def test_non_existent_context_is_valid_for(self, non_existent_context, valid_nest_case, rule_constructor, valid_dataset):
        """Check Rule returns expected result when checking multiple contexts."""
        rule = rule_constructor(non_existent_context, valid_nest_case)
        assert rule.is_valid_for(valid_dataset) is None

    def test_non_existent_context_is_invalid_for(self, non_existent_context, invalid_nest_case, rule_constructor, invalid_dataset):
        """Check Rule returns expected result when checking multiple contexts."""
        rule = rule_constructor(non_existent_context, invalid_nest_case)
        assert rule.is_valid_for(invalid_dataset) is None

    def test_condition_case_is_True_for_valid_dataset(self, valid_condition_rule, valid_dataset):
        """Check that if a condition is `True`, the rule returns None which is considered equivalent to skipping."""
        assert valid_condition_rule.is_valid_for(valid_dataset) is None

    def test_condition_case_is_True_for_invalid_dataset(self, invalid_condition_rule, invalid_dataset):
        """Check that if a condition is `True`, the rule returns None which is considered equivalent to skipping."""
        assert invalid_condition_rule.is_valid_for(invalid_dataset) is None

    def test_condition_case_is_False_for_valid_dataset(self, rule_constructor, valid_single_context, condition_is_false_valid, valid_dataset):
        """Check that if a condition is `False`, the rule validates normally."""
        rule = rule_constructor(valid_single_context, condition_is_false_valid)
        assert rule.is_valid_for(valid_dataset)

    def test_condition_case_is_False_for_invalid_dataset(self, invalid_condition_rule, invalid_dataset):
        """Check that if a condition is `False`, the rule validates normally.

        Note:
            Using an invalid dataset so expecting Rules to evaluate to `False`.

        """
        assert not invalid_condition_rule.is_valid_for(invalid_dataset)

    @pytest.mark.parametrize("junk_condition", [''] + iati.tests.utilities.generate_test_types(['str'], True))
    def test_uninstantiating_condition_case(self, rule_constructor, valid_single_context, validating_case, junk_condition):
        """Check that a non-permitted condition case will not instantiate."""
        junk_condition_case = deepcopy(validating_case)
        junk_condition_case['condition'] = junk_condition
        with pytest.raises(ValueError):
            rule_constructor(valid_single_context, junk_condition_case)


class TestRuleAtLeastOne(RuleSubclassTestBase):
    """A container for tests relating to RuleAtLeastOne."""

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'atleast_one'

    all_valid_cases = [
        {'paths': ['element1']},  # single path
        {'paths': ['element6/@attribute']},
        {'paths': ['element2', 'element3']},  # multiple paths
        {'paths': ['element7/@attribute', 'element8/@attribute']},
        {'paths': ['element4', 'element4']},  # duplicate paths
        {'paths': ['element9/@attribute', 'element9/@attribute']},
        {'paths': ['element11']}  # multiple identitcal elements
    ]

    uninstantiating_cases = [
        {'paths': []},  # empty path array
        {'paths': ['']},  # path is an empty string
        {'paths': 'element'},  # non-array `paths`
        {'paths': [3]},  # non-string value in path array
        {'paths': ['element', 3]},  # mixed string and non-string values in path array
        {'paths': ['element/@attribute', 3]},
        {},  # empty dictionary
        {'paths': {'element'}},  # dictionary paths
        {'paths': {'element/@attribute'}},
        {'paths': 'element'},  # non-array `paths`
        {'paths': [3]},  # non-string value in path array
        {}  # empty dictionary
    ]

    invalidating_cases = [
        {'paths': ['element1']},  # single path, no matches
        {'paths': ['element9/@attribute']},
        {'paths': ['element7', 'element8']},  # multiple paths, both expected matches missing
        {'paths': ['element13/@attribute', 'element14/@attribute']}
    ]

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted case for validating an XML dataset against RuleAtLeastOne."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleAtLeastOne."""
        return request.param

    @pytest.fixture
    def valid_nest_case(self):
        """Non-permitted case for validating an XML dataset against RuleAtLeastOne in nested context."""
        return {'paths': ['element5', 'element10']}

    @pytest.fixture
    def invalid_nest_case(self):
        """Non-permitted case for validating an XML dataset against RuleAtLeastOne in nested context."""
        return {'paths': ['element2', 'element10/@attribute']}

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_atleastone')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_atleastone')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert 'must be present' in str(rule_instantiating)


class TestRuleDateOrder(RuleSubclassTestBase):
    """A container for tests relating to RuleDateOrder.

    Note:
        RuleDateOrder is special and is implemented in a weird way. It is not a happy duck. Therefore, it has special cases that will instantiate but are not valid, which is not the case with other Rules.

    Todo:
        Test implementation of functionality to ignore function call if `less` or `more` do not return dates.

    """

    instatiating_cases = [
        {'less': 'element', 'more': 'element'},  # both `less` and `more` duplicate xpath
        {'less': 'element/@attribute', 'more': 'element/@attribute'},
        {'less': '2017-07-26T13:19:05.493Z', 'more': 'element'},  # `less` is a string-formatted date
        {'less': '2017-07-26T13:19:05.493Z', 'more': 'element/@attribute'},
        {'less': 'element', 'more': '2017-07-26T13:19:05.493Z'},  # `more` is a string-formatted date
        {'less': 'element/@attribute', 'more': '2017-07-26T13:19:05.493Z'},
        {'less': 'NOW', 'more': 'NOW'}  # both `less` and `more` as NOW
    ]

    validating_cases = [
        {'less': 'element1', 'more': 'element2'},  # correct date order
        {'less': 'element11/@attribute', 'more': 'element12/@attribute'},
        {'less': 'element3', 'more': 'NOW'},  # `more` is NOW
        {'less': 'element13/@attribute', 'more': 'NOW'},
        {'less': 'NOW', 'more': 'element4'},  # `less` is NOW
        {'less': 'NOW', 'more': 'element14/@attribute'},
        {'less': 'element5', 'more': 'element6'},  # multiple identical `less` values
        {'less': 'element15/@attribute', 'more': 'element16/@attribute'},
        {'less': 'element7', 'more': 'element8'},  # multiple identical `more` values
        {'less': 'element17/@attribute', 'more': 'element18/@attribute'},
        {'less': '//element9', 'more': '//element10'},  # nested data
        {'less': '//element19/@attribute', 'more': '//element20/@attribute'},
        {'less': 'element21', 'more': 'element22'},  # positive timezone value
        {'less': 'element29/@attribute', 'more': 'element30/@attribute'},
        {'less': 'element23', 'more': 'element24'},  # negative timezone value
        {'less': 'element31/@attribute', 'more': 'element32/@attribute'},
        {'less': 'element25', 'more': 'element26'},  # timezone not on the hour
        {'less': 'element33/@attribute', 'more': 'element34/@attribute'},
        {'less': 'element27', 'more': 'element28'},  # UTC timezone
        {'less': 'element35/@attribute', 'more': 'element36/@attribute'},
        {'less': 'nOw', 'more': 'noW'},  # not special case, should treat as regular path value
        {'less': 'now/@attribute', 'more': 'Now/@attribute'},
        {'less': 'element37', 'more': 'element38'},  # multiple identical elements
        {'less': 'element47', 'more': 'element48'},  # dates have identical values
        {'less': 'element49/@attribute', 'more': 'element50/@attribute'},
        {'less': 'element51', 'more': 'element51'},  # `less` and `more` reference the same date
        {'less': 'element52/@attribute', 'more': 'element52/@attribute'}
    ]

    all_valid_cases = instatiating_cases + validating_cases

    uninstantiating_cases = [
        {'less': 'start'},  # missing required attribute - `more`
        {'more': 'end'},  # missing required attribute - `less`
        {'less': '', 'more': ''},  # path case is empty string
        {'less': 1501075031590, 'more': 'end-xpath'},  # `less` is not a string
        {'less': 'start-xpath', 'more': 1501075031590},  # `more` is not a string
        {},  # empty dictionary
        {'less': ['start']},  # less is a list
        {'more': ['end']},  # more is a list
    ]

    invalidating_cases = [
        {'less': 'element1', 'more': 'element2'},  # dates not in chronological order
        {'less': 'element14/@attribute', 'more': 'element15/@attribute'},
        {'less': 'NOW', 'more': 'element5'},  # `more` is chronologically before NOW
        {'less': 'NOW', 'more': 'element18/@attribute'},
        {'less': 'element6', 'more': 'NOW'},  # `less` is chronologically after NOW
        {'less': 'element19/@attribute', 'more': 'NOW'},
        {'less': '//element7', 'more': '//element8'},  # nested dates not in chronological order
        {'less': '//element20/@attribute', 'more': '//element21/@attribute'},
        {'less': 'element10', 'more': 'element11'},  # multiple identical `less` dates that are chronologically after `more`
        {'less': 'element23/@attribute', 'more': 'element24/@attribute'},
        {'less': 'element12', 'more': 'element13'},  # multiple identical `more` dates that are chronologically before `less`
        {'less': 'element25/@attribute', 'more': 'element26/@attribute'},
        {'less': 'element27', 'more': 'element28'},  # multiple identical elements in incorrect order
        {'less': 'element29', 'more': 'xpath-that-does-not-exist'},  # the xpath for more does not exist
        {'less': 'xpath-that-does-not-exist', 'more': 'element29'}  # the xpath for less does not exist
    ]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'date_order'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=validating_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleDateOrder."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleDateOrder."""
        return request.param

    @pytest.fixture(params=[
        {'less': '//element9', 'more': '//element10'},
        {'less': '//element19/@attribute', 'more': '//element20/@attribute'}
    ])
    def valid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleDateOrder in nested context."""
        return request.param

    @pytest.fixture(params=[
        {'less': '//element7', 'more': '//element8'},
        {'less': '//element20/@attribute', 'more': '//element21/@attribute'}
    ])
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleDateOrder in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for instatiating this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_dateorder')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_dateorder')

    @pytest.mark.parametrize("case", [
        {'less': 'element1', 'more': 'element2'},  # Euro-date format
        {'less': 'element17/@attribute', 'more': 'element18/@attribute'},
        {'less': 'element3', 'more': 'element4'},  # USA-date format
        {'less': 'element19/@attribute', 'more': 'element20/@attribute'},
        {'less': 'element5', 'more': 'element6'},  # month is given as text
        {'less': 'element21/@attribute', 'more': 'element22/@attribute'},
        {'less': 'element7', 'more': 'element8'},  # year given in shortened format
        {'less': 'element23/@attribute', 'more': 'element24/@attribute'},
        {'less': 'element9', 'more': 'element10'},  # leading characters included
        {'less': 'element25/@attribute', 'more': 'element26/@attribute'},
        {'less': 'element11', 'more': 'element12'},  # non-permitted trailing characters
        {'less': 'element27/@attribute', 'more': 'element28/@attribute'},
        {'less': 'element13', 'more': 'element14'},  # leading whitespace
        {'less': 'element29/@attribute', 'more': 'element30/@attribute'},
        {'less': 'element15', 'more': 'element16'},  # trailing whitespace
        {'less': 'element31/@attribute', 'more': 'element32/@attribute'},
        {'less': 'element33', 'more': 'element34'},  # timezone not zero-padded
        {'less': 'element35/@attribute', 'more': 'element36/@attribute'},
        {'less': 'element37', 'more': 'element38'},  # non-permitted leading timezone character
        {'less': 'element39/@attribute', 'more': 'element40/@attribute'},
        {'less': 'element41', 'more': 'element42'},  # multiple identical elements but non-duplicated values
        {'less': 'element43', 'more': 'element44'},  # UNIX timestamp format
        {'less': 'element45/@attribute', 'more': 'element46/@attribute'},
        {'less': 'element47', 'more': 'element48'},  # All text date format
        {'less': 'element49/@attribute', 'more': 'element50/@attribute'},
        {'less': 'element51', 'more': 'element52'},  # missing day
        {'less': 'element53', 'more': 'element54'},  # day not zero-padded
        {'less': 'element55', 'more': 'element56'}  # month not zero-padded
    ])
    def test_incorrect_date_format_raises_error(self, valid_single_context, case, rule_constructor):
        """Check that a dataset with dates in an incorrect format raise expected error."""
        rule = rule_constructor(valid_single_context, case)
        with pytest.raises(ValueError):
            rule.is_valid_for(iati.tests.utilities.load_as_dataset('ruleset/invalid_format_dateorder'))

    @pytest.mark.parametrize("case", [
        {'less': 'element39', 'more': 'element40'},  # `less` date missing
        {'less': 'element43', 'more': 'element44'},
        {'less': 'element41', 'more': 'element42'},  # `more` date missing
        {'less': 'element45', 'more': 'element46'}
    ])
    def test_is_valid_for_returns_None(self, valid_single_context, case, rule_constructor, valid_dataset):
        """Check that None is returned when expected date not found."""
        rule = rule_constructor(valid_single_context, case)
        assert rule.is_valid_for(valid_dataset) is None

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert any(needle in str(rule_instantiating) for needle in ['must be chronologically', 'in the future', 'in the past'])


class TestRuleDependent(RuleSubclassTestBase):
    """A container for tests relating to RuleDependent."""

    all_valid_cases = [
        {'paths': ['element1']},  # single path
        {'paths': ['element7/@attribute']},
        {'paths': ['element2', 'element3']},  # multiple paths
        {'paths': ['element8/@attribute', 'element9/@attribute']},
        {'paths': ['element4', 'element4']},  # duplicate paths
        {'paths': ['element10/@attribute', 'element10/@attribute']},
        {'paths': ['element13', 'element14']}  # duplicate elements
    ]

    uninstantiating_cases = [
        {'paths': []},  # empty path array
        {'paths': ['']},  # path is an empty string
        {'paths': 'path_1'},  # non-array `paths`
        {'paths': [3]},  # non-string value in path array
        {'paths': ['path_1', 3]},  # mixed string and non-string value in path array
        {},  # empty dictionary
        {'paths': {'path_1'}}  # path value is dictionary instead of list
    ]

    invalidating_cases = [
        {'paths': ['element1', 'element2', 'element3']},  # dependent element missing
        {'paths': ['element4/@attribute', 'element5/@attribute', 'element6/@attribute']},  # dependent attribute missing
        {'paths': ['element10', 'element11']}  # dependent element of duplicate element missing
    ]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'dependent'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleDependent."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleDependent."""
        return request.param

    @pytest.fixture(params=[
        {'paths': ['./element5', './element6']},
        {'paths': ['./element11/@attribute', './element12/@attribute']}

    ])
    def valid_nest_case(self, request):
        """Permitted case for validating an XML dataset against RuleDependent in nested context."""
        return request.param

    @pytest.fixture(params=[
        {'paths': ['./element7', './element8']},
        {'paths': ['./element9/@attribute', './element10/@attribute']},
        {'paths': ['./element15', './element16']}
    ])
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleDependent in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for instatiating this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_dependent')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_dependent')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert any(needle in str(rule_instantiating) for needle in ['must all exist', 'always True'])


class TestRuleNoMoreThanOne(RuleSubclassTestBase):
    """A container for tests relating to RuleNoMoreThanOne."""

    all_valid_cases = [
        {'paths': ['element1']},  # single path
        {'paths': ['element5/@attribute']},
        {'paths': ['element2', 'element3']},  # multiple paths
        {'paths': ['element6/@attribute', 'element7/@attribute']},
        {'paths': ['element4', 'element4']},  # duplicate paths
        {'paths': ['element8/@attribute', 'element8/@attribute']}
    ]

    uninstantiating_cases = [
        {'paths': []},  # empty path array
        {'paths': ['']},  # path is an empty string
        {'paths': 'path_1'},  # non-array `paths`
        {'paths': [3]},  # non-string value in path array
        {'paths': ['path_1', 3]},  # mixed string and non-string value in path array
        {},  # empty dictionary
        {'paths': {'path_1'}}  # dictionary paths
    ]

    invalidating_cases = [
        {'paths': ['element1']},  # single path
        {'paths': ['element7/@attribute']},
        {'paths': ['element2', 'element3']},  # multiple paths, multiple of each
        {'paths': ['element8/@attribute', 'element9/@attribute']},
        {'paths': ['element4', 'element4']},  # duplicate paths
        {'paths': ['element10/@attribute', 'element10/@attribute']},
        {'paths': ['element5', 'element6']},  # one path valid, another invalid
        {'paths': ['element11/@attribute', 'element12/@attribute']},
        {'paths': ['element15', 'element16']},  # multiple paths, one of each
        {'paths': ['element17/@attribute', 'element18/@attribute']}  # multiple paths, one of each
    ]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'no_more_than_one'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleNoMoreThanOne."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted casesd for validating an XML dataset against RuleNoMoreThanOne."""
        return request.param

    @pytest.fixture
    def valid_nest_case(self):
        """Permitted case for validating an XML dataset against RuleNoMoreThanOne in nested context."""
        return {'paths': ['element9', 'element10/@attribute']}

    @pytest.fixture
    def invalid_nest_case(self):
        """Non-permitted case for validating an XML dataset against RuleNoMoreThanOne in nested context."""
        return {'paths': ['element13', 'element14/@attribute']}

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_nomorethanone')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_nomorethanone')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert any(needle in str(rule_instantiating) for needle in ['zero or one', 'no more than one'])


class TestRuleRegexMatches(RuleSubclassTestBase):
    """A container for tests relating to RuleRegexMatches."""

    all_valid_cases = [
        {'regex': r'\btest\b', 'paths': ['element1']},  # single path with regex
        {'regex': r'\btest\b', 'paths': ['element5/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element2', 'element3']},  # multiple paths with regex
        {'regex': r'\btest\b', 'paths': ['element6/@attribute', 'element7/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element4', 'element4']},  # duplicate paths with regex
        {'regex': r'\btest\b', 'paths': ['element8/@attribute', 'element8/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element11']},  # duplicate element
        {'regex': r'\btest\b', 'paths': ['element12']},  # no such element exists
        {'regex': r'\btest\b', 'paths': ['element13/@attribute']}  # no such attribute exists
    ]

    uninstantiating_cases = [
        {'regex': r'some regex', 'paths': []},  # empty path array
        {'regex': r'some regex', 'paths': ['']},  # paths is an empty string
        {'regex': r'some regex', 'paths': 'path_1'},  # non-array `paths`
        {'regex': r'some regex', 'paths': [3]},  # non-string value in path array
        {'regex': r'some regex', 'paths': ['path_1', 3]},  # mixed string and non-string value in path array
        {'regex': r'some regex'},  # missing required attribute - `paths`
        {'paths': ['path_1', 'path_2']},  # missing required attribute - `regex`
        {'regex': r'[', 'paths': ['path_1']},  # provided string not a valid regex
        {'regex': 3, 'paths': ['path_1']},  # provided regex not a string
        {},  # empty dictionary
        {'regex': r'some regex', 'paths': {'path_1'}},  # dictionary paths
        {'regex': [r'some regex'], 'paths': 'path_1'},  # list regex
        {'regex': r'', 'paths': ['path_1']}  # regex is an empty string
    ]

    invalidating_cases = [
        {'regex': r'\btest\b', 'paths': ['element1']},  # single path with regex
        {'regex': r'\btest\b', 'paths': ['element5/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element2', 'element3']},  # multiple paths with regex
        {'regex': r'\btest\b', 'paths': ['element6/@attribute', 'element7/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element4', 'element4']},  # duplicate paths with regex
        {'regex': r'\btest\b', 'paths': ['element8/@attribute', 'element8/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element11']}  # duplicate element
    ]

    nest_case = [{'regex': r'\btest\b', 'paths': ['./element9', './element10/@attribute']}]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'regex_matches'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleRegexMatches."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleRegexMatches."""
        return request.param

    @pytest.fixture(params=nest_case)
    def valid_nest_case(self, request):
        """Permitted case for validating an XML dataset against RuleRegexMatches in nested context."""
        return request.param

    @pytest.fixture(params=nest_case)
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleRegexMatches in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_regexmatches')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_regexmatches')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert 'must match the regular expression' in str(rule_instantiating)


class TestRuleRegexNoMatches(RuleSubclassTestBase):
    """A container for tests relating to RuleRegexNoMatches."""

    all_valid_cases = [
        {'regex': r'\btest\b', 'paths': ['element1']},  # single path with regex
        {'regex': r'\btest\b', 'paths': ['element5/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element2', 'element3']},  # multiple paths with regex
        {'regex': r'\btest\b', 'paths': ['element6/@attribute', 'element7/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element4', 'element4']},  # duplicate paths with regex
        {'regex': r'\btest\b', 'paths': ['element8/@attribute', 'element8/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element11']},  # duplicate element
        {'regex': r'\btest\b', 'paths': ['element12']},  # no such element exists
        {'regex': r'\btest\b', 'paths': ['element13/@attribute']}  # no such attribute exists
    ]

    uninstantiating_cases = [
        {'regex': 'some regex', 'paths': []},  # empty path array
        {'regex': r'some regex', 'paths': ['']},  # paths is an empty string
        {'regex': 'some regex', 'paths': 'path_1'},  # non-array `paths`
        {'regex': 'some regex', 'paths': [3]},  # non-string value in path array
        {'regex': 'some regex', 'paths': ['path_1', 3]},  # mixed string and non-string value in path array
        {'regex': 'some regex'},  # missing required attribute - `paths`
        {'paths': ['path_1', 'path_2']},  # missing required attribute - `regex`
        {'regex': '[', 'paths': ['path_1']},  # provided string not a valid regex
        {'regex': 3, 'paths': ['path_1']},  # provided regex not a string
        {},  # empty dictionary
        {'regex': 'some regex', 'paths': {'path_1'}},  # dictionary paths
        {'regex': ['some regex'], 'paths': 'path_1'},  # list regex
        {'regex': '', 'paths': ['path_1']}  # regex is an empty string
    ]

    invalidating_cases = [
        {'regex': r'\btest\b', 'paths': ['element1']},  # single path with regex
        {'regex': r'\btest\b', 'paths': ['element5/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element2', 'element3']},  # multiple paths with regex
        {'regex': r'\btest\b', 'paths': ['element6/@attribute', 'element7/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element4', 'element4']},  # duplicate paths with regex
        {'regex': r'\btest\b', 'paths': ['element8/@attribute', 'element8/@attribute']},
        {'regex': r'\btest\b', 'paths': ['element11']}  # duplicate element
    ]

    nest_case = [{'regex': r'\btest\b', 'paths': ['./element9', './element10/@attribute']}]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'regex_no_matches'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleRegexNoMatches."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleRegexNoMatches."""
        return request.param

    @pytest.fixture(params=nest_case)
    def valid_nest_case(self, request):
        """Permitted case for validating an XML dataset against RuleRegexNoMatches in nested context."""
        return request.param

    @pytest.fixture(params=nest_case)
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleRegexNoMatches in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_regexnomatches')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_regexnomatches')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert 'must not match the regular expression' in str(rule_instantiating)


class TestRuleStartsWith(RuleSubclassTestBase):
    """A container for tests relating to RuleStartsWith."""

    all_valid_cases = [
        {'start': 'prefix', 'paths': ['element1']},  # single path with valid prefix string
        {'start': 'prefix', 'paths': ['element5/@attribute']},
        {'start': 'prefix', 'paths': ['element2', 'element3']},  # multiple paths with valid prefix string
        {'start': 'prefix', 'paths': ['element6/@attribute', 'element7/@attribute']},
        {'start': 'prefix', 'paths': ['element4', 'element4']},  # duplicate paths with valid prefix string
        {'start': 'prefix', 'paths': ['element8/@attribute', 'element8/@attribute']},
        {'start': 'prefix', 'paths': ['element11']},  # duplicate element
        {'start': 'prefix', 'paths': ['element12']},  # no such element exists
        {'start': 'prefix', 'paths': ['element13/@attribute']}  # no such attribute exists
    ]

    uninstantiating_cases = [
        {'start': 'prefix-xpath', 'paths': []},  # empty path array
        {'start': 'prefix-xpath', 'paths': 'path_1'},  # non-array `paths`
        {'start': 'prefix-xpath', 'paths': [3]},  # non-string value in path array
        {'start': 'prefix-xpath', 'paths': ['path_1', 3]},  # mixed string and non-string value in path array
        {'start': 3, 'paths': ['path_1']},  # provided prefix xpath not a string
        {'start': 'prefix-xpath'},  # missing required attribute - `paths`
        {'paths': ['path_1', 'path_2']},  # missing required attribute - `start`
        {},  # empty dictionary
        {'start': 'prefix-xpath', 'paths': {'path_1'}},  # dictionary paths
        {'start': ['prefix-xpath'], 'paths': ['path_1']},  # list start
        {'start': '', 'paths': ['path_1']},  # start is an empty string
        {'start': 'prefix_path', 'paths': ['']}  # paths is an empty string
    ]

    invalidating_cases = [
        {'start': 'prefix', 'paths': ['element1']},  # single path with valid prefix string
        {'start': 'prefix', 'paths': ['element5/@attribute']},
        {'start': 'prefix', 'paths': ['element2', 'element3']},  # multiple paths with valid prefix string
        {'start': 'prefix', 'paths': ['element6/@attribute', 'element7/@attribute']},
        {'start': 'prefix', 'paths': ['element4', 'element4']},  # duplicate paths with valid prefix string
        {'start': 'prefix', 'paths': ['element8/@attribute', 'element8/@attribute']},
        {'start': 'prefix', 'paths': ['element11']}  # duplicate element
    ]

    nest_case = [{'start': '//prefix', 'paths': ['element9', 'element10/@attribute']}]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'startswith'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleStartsWith."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleStartsWith."""
        return request.param

    @pytest.fixture(params=nest_case)
    def valid_nest_case(self, request):
        """Permitted case for validating an XML dataset against RuleStartsWith in nested context."""
        return request.param

    @pytest.fixture(params=nest_case)
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleStartsWith in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_startswith')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_startswith')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert 'must start with' in str(rule_instantiating)

    @pytest.mark.parametrize("test_case", [
        {'start': 'duplicateprefix', 'paths': ['element12']},  # `start` matches multiple elements with the same text value
        {'start': 'multiprefix', 'paths': ['element13']}  # `start` matches multiple elements with different text values
    ])
    def test_multiple_matching_start_elements_raise_error(self, valid_single_context, rule_constructor, test_case, invalid_dataset):
        """Check that an error is raised when the specified XPath for `start` returns multiple elements."""
        rule = rule_constructor(valid_single_context, test_case)
        with pytest.raises(ValueError):
            rule.is_valid_for(invalid_dataset)

    def test_missing_start_value_raises_error(self, valid_single_context, rule_constructor, invalid_dataset):
        """Check that if no prefix value is found the rule returns None which is considered equivalent to skipping."""
        missing_value_case = {'start': 'missingprefix', 'paths': ['element14']}
        rule = rule_constructor(valid_single_context, missing_value_case)
        with pytest.raises(ValueError):
            rule.is_valid_for(invalid_dataset)


class TestRuleSum(RuleSubclassTestBase):
    """A container for tests relating to RuleSum.

    Todo:
        **Determine if assumption that double counting of elements should be not permitted when duplicate paths specified, but should when multiple elements exist, is correct.

    """

    all_valid_cases = [
        {'paths': ['element1'], 'sum': 3},  # single path with sum
        {'paths': ['element19/@attribute'], 'sum': 3},
        {'paths': ['element2', 'element3'], 'sum': 3},  # multiple paths with sum
        {'paths': ['element20/@attribute', 'element21/@attribute'], 'sum': 3},
        {'paths': ['element4', 'element4'], 'sum': 3},  # duplicate paths with sum **
        {'paths': ['element22/@attribute', 'element22/@attribute'], 'sum': 3},  # **
        {'paths': ['element5', 'element6'], 'sum': -1000},  # negative sum
        {'paths': ['element23/@attribute', 'element24/@attribute'], 'sum': -1000},
        {'paths': ['element7', 'element8'], 'sum': 101},  # sum greater than standard percentage limit
        {'paths': ['element25/@attribute', 'element26/@attribute'], 'sum': 101},
        {'paths': ['element9', 'element10'], 'sum': 15.5},  # decimal sum
        {'paths': ['element27/@attribute', 'element28/@attribute'], 'sum': 15.5},
        {'paths': ['element11', 'element12'], 'sum': 0},  # zero sum
        {'paths': ['element29/@attribute', 'element30/@attribute'], 'sum': 0},
        {'paths': ['element13', 'element14'], 'sum': float(10**100)},  # big sum
        {'paths': ['element31/@attribute', 'element32/@attribute'], 'sum': float(10**100)},
        {'paths': ['element15', 'element16'], 'sum': float(-10**100)},  # tiny sum
        {'paths': ['element33/@attribute', 'element34/@attribute'], 'sum': float(-10**100)},
        {'paths': ['element17', 'element18'], 'sum': 2.99792458e6},  # exponential sum
        {'paths': ['element35/@attribute', 'element36/@attribute'], 'sum': 2.99792458e6},
        {'paths': ['element37'], 'sum': 50},  # duplicate elements in data **
        {'paths': ['element42', 'element43'], 'sum': 0.3},  # sum to value that cannot be represented using standard binary representation
        {'paths': ['element44/@attribute', 'element45/@attribute'], 'sum': 0.3},
        {'paths': ['element46'], 'sum': 50}  # whitespace around numeric value
    ]

    uninstantiating_cases = [
        {'paths': [], 'sum': 3},  # empty path array
        {'paths': 'path_1', 'sum': 3},  # non-array `paths`
        {'paths': [3], 'sum': 3},  # non-string value in path array
        {'paths': ['path_1', 3], 'sum': 3},  # mixed string and non-string value in path array
        {'paths': ['path_1', 'path_2']},  # missing required attribute - `sum`
        {'sum': 100},  # missing required attribute - `paths`
        {},  # empty dictionary
        {'paths': ['path_1'], 'sum': '3'},  # sum is a string representation of a number
        {'sum': 100, 'paths': {'path_1'}},  # dictionary paths
        {'sum': [100], 'paths': 'path_1'},  # list sum
        {'paths': [''], 'sum': 100},  # empty paths string
        {'paths': ['path_1'], 'sum': ''}  # sum is empty string
    ]

    invalidating_cases = [
        {'paths': ['element1'], 'sum': 3},  # single path with sum
        {'paths': ['element19/@attribute'], 'sum': 3},
        {'paths': ['element2', 'element3'], 'sum': 3},  # multiple paths with sum
        {'paths': ['element20/@attribute', 'element21/@attribute'], 'sum': 3},
        {'paths': ['element4', 'element4'], 'sum': 3},  # duplicate paths with sum **
        {'paths': ['element22/@attribute', 'element22/@attribute'], 'sum': 3},  # **
        {'paths': ['element5', 'element6'], 'sum': -1000},  # negative sum
        {'paths': ['element23/@attribute', 'element24/@attribute'], 'sum': -1000},
        {'paths': ['element7', 'element8'], 'sum': 101},  # sum greater than standard percentage limit
        {'paths': ['element25/@attribute', 'element26/@attribute'], 'sum': 101},
        {'paths': ['element9', 'element10'], 'sum': 15.5},  # decimal sum
        {'paths': ['element27/@attribute', 'element28/@attribute'], 'sum': 15.5},
        {'paths': ['element11', 'element12'], 'sum': 0},  # zero sum
        {'paths': ['element29/@attribute', 'element30/@attribute'], 'sum': 0},
        {'paths': ['element13', 'element14'], 'sum': float(10**100)},  # big sum
        {'paths': ['element31/@attribute', 'element32/@attribute'], 'sum': float(10**100)},
        {'paths': ['element15', 'element16'], 'sum': float(-10**100)},  # tiny sum
        {'paths': ['element33/@attribute', 'element34/@attribute'], 'sum': float(-10**100)},
        {'paths': ['element17', 'element18'], 'sum': 2.99792458e6},  # exponential sum
        {'paths': ['element35/@attribute', 'element36/@attribute'], 'sum': 2.99792458e6},
        {'paths': ['element37'], 'sum': 50}  # duplicate elements in data **
    ]

    nest_cases = [
        {'paths': ['//element38', '//element39'], 'sum': 3},
        {'paths': ['//element40/@attribute', '//element41/@attribute'], 'sum': 3}
    ]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'sum'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleSum."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleSum."""
        return request.param

    @pytest.fixture(params=nest_cases)
    def valid_nest_case(self, request):
        """Permitted case for validating an XML dataset against RuleSum in nested context."""
        return request.param

    @pytest.fixture(params=nest_cases)
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleSum in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_sum')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_sum')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert 'sum of values' in str(rule_instantiating)

    @pytest.mark.parametrize("test_case", [
        {'paths': ['element42'], 'sum': 50}  # non-numeric value
    ])
    def test_non_numeric_value_raise_error(self, valid_single_context, rule_constructor, test_case, invalid_dataset):
        """Check that an error is raised when the specified XPath for `start` returns multiple elements."""
        rule = rule_constructor(valid_single_context, test_case)
        with pytest.raises(ValueError):
            rule.is_valid_for(invalid_dataset)

    def test_no_values_to_sum_skips(self, valid_single_context, rule_constructor, valid_dataset):
        """Check that the Rule is skipped if no values are found to compare to the value of `sum`."""
        no_values_case = {'paths': ['this_element_does_not_exist', 'neither_does_this_one'], 'sum': 100}
        rule = rule_constructor(valid_single_context, no_values_case)
        assert rule.is_valid_for(valid_dataset) is None


class TestRuleUnique(RuleSubclassTestBase):
    """A container for tests relating to RuleUnique."""

    all_valid_cases = [
        {'paths': ['element1']},  # single path
        {'paths': ['element5/@attribute']},
        {'paths': ['element2', 'element3']},  # multiple paths
        {'paths': ['element6/@attribute', 'element7/@attribute']},
        {'paths': ['element4', 'element4']},  # duplicate paths
        {'paths': ['element8/@attribute', 'element8/@attribute']},
        {'paths': ['element13']},  # duplicate elements exist for path
        {'paths': ['element14']},  # no such element exists
        {'paths': ['element15/@attribute']}  # no such attribute exists
    ]

    uninstantiating_cases = [
        {'paths': []},  # empty path array
        {'paths': 'path_1'},  # non-array `paths`
        {'paths': [3]},  # non-string value in path array
        {'paths': ['path_1', 3]},  # mixed string and non-string value in path array
        {},  # empty dictionary
        {'paths': {'path_1'}}
    ]

    invalidating_cases = [
        {'paths': ['element1', 'element2']},  # multiple paths
        {'paths': ['element3/@attribute', 'element4/@attribute']},
        {'paths': ['element9']}  # duplicate elements exist for path
    ]

    @pytest.fixture
    def rule_type(self):
        """Type of Rule."""
        return 'unique'

    @pytest.fixture(params=all_valid_cases)
    def instantiating_case(self, request):
        """Permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=uninstantiating_cases)
    def uninstantiating_case(self, request):
        """Non-permitted case for instatiating this Rule."""
        return request.param

    @pytest.fixture(params=all_valid_cases)
    def validating_case(self, request):
        """Permitted cases for validating an XML dataset against RuleUnique."""
        return request.param

    @pytest.fixture(params=invalidating_cases)
    def invalidating_case(self, request):
        """Non-permitted cases for validating an XML dataset against RuleUnique."""
        return request.param

    @pytest.fixture(params=[
        {'paths': ['element9', 'element10']},
        {'paths': ['element11/@attribute', 'element12/@attribute']}
    ])
    def valid_nest_case(self, request):
        """Permitted case for validating an XML dataset against RuleUnique in nested context."""
        return request.param

    @pytest.fixture(params=[
        {'paths': ['element5', 'element6']},
        {'paths': ['element7/@attribute', 'element8/@attribute']}
    ])
    def invalid_nest_case(self, request):
        """Non-permitted case for validating an XML dataset against RuleUnique in nested context."""
        return request.param

    @pytest.fixture
    def invalid_dataset(self):
        """Invalid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/invalid_unique')

    @pytest.fixture
    def valid_dataset(self):
        """Return valid dataset for this Rule."""
        return iati.tests.utilities.load_as_dataset('ruleset/valid_unique')

    def test_rule_string_output_specific(self, rule_instantiating):
        """Check that the string format of the Rule contains some relevant information."""
        assert 'must be unique' in str(rule_instantiating)



"""A module containing utility constants and functions for tests.

A large number of constants containing example file content are contained. These constants are named from left to right, with general properties first, then leading into more specific information. These names indicate what they are used for.

Example:
    To load a file into a string::

        name_of_file = 'a-file-name-without-the-extension'
        CONSTANT_NAME = load_as_string(name_of_file)

Note:
    The current method of managing test data is known to be sub-optimal. Suggestions for better methods that satisfy requirements are appreciated!

Todo:
    Add versions of constants that are valid for differing schema versions.

"""
import decimal
from lxml import etree
import iati.resources
import iati.constants


def load_as_dataset(file_path):
    """Load a specified test data file as a Dataset.

    Args:
        file_path (str): The path of the file, relative to the root test data folder. Folders should be separated by a forward-slash (`/`).

    Returns:
        dataset: A Dataset containing the contents of the file at the specified location.

    Raises:
        iati.exceptions.ValidationError: If the provided XML does not conform to the IATI standard.

    """
    return iati.resources.load_as_dataset(iati.resources.get_test_data_path(file_path))


def load_as_string(file_path):
    """Load a specified test data file as a string.

    Args:
        file_path (str): The path of the file, relative to the root test data folder. Folders should be separated by a forward-slash (`/`).

    Returns:
        str (python3) / unicode (python2): The contents of the file at the specified location.

    """
    return iati.resources.load_as_string(iati.resources.get_test_data_path(file_path))


# This will need updating once test data forcing XML format is fixed
RULESET_FOR_TESTING = iati.Ruleset(iati.resources.load_as_string(iati.resources.get_ruleset_path('ruleset_for_tests')))
"""A working Ruleset based on the Standard Ruleset."""


SCHEMA_ACTIVITY_NAME_VALID = 'iati-activities-schema'
"""A string containing a valid IATI Activity Schema name."""
SCHEMA_ORGANISATION_NAME_VALID = 'iati-organisations-schema'
"""A string containing a valid IATI Organisaion Schema name."""
SCHEMA_NAME_VALID = 'iati-activities-schema'
"""A string containing a valid Schema name."""

XML_TREE_VALID = etree.fromstring(load_as_string('valid_not_iati'))
"""An etree that is valid XML but not IATI XML."""
XML_TREE_VALID_IATI = etree.fromstring(load_as_string('valid_iati'))
"""A valid IATI etree."""
XML_TREE_VALID_IATI_INVALID_CODE = etree.fromstring(load_as_string('valid_iati_invalid_code'))
"""A valid IATI etree that has an invalid Code value."""


TYPE_TEST_DATA = {
    'bool': [True, False],
    'bytes': [],  # counts as a string, so moved there
    'bytearray': [bytearray.fromhex('2Ef0 F1f2  '), bytearray(b'Hi!'), bytearray(range(20))],
    'complex': [3453J, -35415J, 0J, complex(234, 681), complex(-768, 16078), complex(6187, -81), complex(-1867, -618)],
    'contextmanager': [decimal.localcontext()],
    'float': [-2343324.534, 0.0, 32423.34, 1.3e7, float('Infinity'), float('nan'), float('-inf')],
    'function': [dir, hash, map, max, type],
    'int': [-234, 0, 2131908],
    'iter': [iter([1, 2, 3])],
    'list': [[], [1, 23], ['varying', 7, b'ypes', []]],
    'mapping': [{}, dict(zip(['one', 'two', 'three'], [1, 2, 3])), dict(one=1, two=2, three=3)],
    'memory': [memoryview(b'abcefg')],
    'none': [None],
    'other': [NotImplemented],
    'range': [range(3, 4)],
    'set': [set(range(20)), set(['hello', 23]), frozenset(range(20)), frozenset(['hello', 23])],
    'str': [b'\x80abc', b'\x80abc', '\N{GREEK CAPITAL LETTER DELTA}', '\u0394', '\U00000394', 'This is a string'],  # python2.7 warning # pylint: disable=anomalous-unicode-escape-in-string
    'tuple': [(), (1, 2)],
    'type': [type(1), type('string')],
    'unicode': [],  # counts as a string, so moved there
    'view': [{}.keys(), dict(zip(['one', 'two', 'three'], [1, 2, 3])).items(), dict(one=1, two=2, three=3).values()]
}
"""Generic test data of various Python builtin types."""


def generate_test_types(types, invert_types=False):
    """Find a number of values of the specified type to pass to a test function.

    Args:
        types (list of str): The types of parameter that should be looked for.
        invert_types (bool): Whether to invert the list of types being looked for, instead returning everything else. Default False.

    Returns:
        list: A list of values to pass to the test function.

    """
    valid_keys_as_specified = [key for key in types if key in TYPE_TEST_DATA]
    if invert_types:
        valid_keys = [key for key in TYPE_TEST_DATA if key not in valid_keys_as_specified]
    else:
        valid_keys = valid_keys_as_specified

    valid_keys = sorted(valid_keys)

    results = []

    for key in valid_keys:
        results = results + TYPE_TEST_DATA[key]

    return results



"""A module containing a core representation of an IATI Dataset."""
import sys
from lxml import etree
import iati.exceptions
import iati.utilities
import iati.validator


class Dataset(object):
    """Representation of an IATI XML file that may be validated against a Schema.

    Attributes:
        xml_str (str): An XML string representation of the Dataset.
        xml_tree (ElementTree): A tree representation of the Dataset.

    Note:
        Should it be modified after initialisation, the current content of the Dataset is deemed to be that which was last asigned to either `self.xml_str` or `self.xml_tree`.

    Warning:
        The behaviour of simultaneous assignment to both `self.xml_str` and `self.xml_tree` is undefined.

        Does not fully hide the lxml internal workings.

    Todo:
        `xml_str` and `xml_tree` are not great names. They are also too tied together. It should be determined whether this close relationship is really desired.

        Implement a number of helper functions for common operations.

        Implement getters and setters for attributes.

        Implement an addition override to allow for combation of Datasets.

    """

    def __init__(self, xml):
        """Initialise a Dataset.

        Args:
            xml (str or ElementTree): A representation of the XML to encapsulate.
                May be either a string or a lxml ElementTree.

        Raises:
            TypeError: If an attempt to pass something that is not a string or ElementTree is made.
            ValueError: If a provided XML string is not valid XML.

        Warning:
            The required parameters to create a Dataset may change. See the TODO.

        Todo:
            It should be possible to create a Dataset from a file. In this situation, having `xml` as a required parameter does not seem sensible. Need to better consider this situation.

            Add a way to determine whether a Dataset fully conforms to the IATI Standard and / or modify the Dataset so that it does.

        """
        self._xml_str = None
        self._xml_tree = None

        if isinstance(xml, etree._Element):  # pylint: disable=W0212
            self.xml_tree = xml
        else:
            self.xml_str = xml

    @property
    def xml_str(self):
        """str: An XML string representation of the Dataset.

        Raises:
            ValueError: If a value that is being assigned is not a valid XML string.
            TypeError: If a value that is being assigned is not a string.

        Todo:
            Clarify error messages, for example when a mismatched encoding is used.

            Perhaps pass on the original lxml error message instead of trying to intrepret what might have gone wrong when running `etree.fromstring()`.

        """
        return self._xml_str

    @xml_str.setter
    def xml_str(self, value):
        if isinstance(value, etree._Element):  # pylint: disable=W0212
            msg = "If setting a Dataset with an ElementTree, use the xml_tree property, not the xml_str property."
            iati.utilities.log_error(msg)
            raise TypeError(msg)
        else:
            try:
                value_stripped = value.strip()

                # Convert the input to bytes, as etree.fromstring works most consistently with bytes objects, especially if an XML encoding declaration has been used.
                if (isinstance(value_stripped, str) and
                        sys.version_info.major > 2):  # Python v2 treats strings as byte objects by default
                    value_stripped_bytes = value_stripped.encode()
                else:
                    value_stripped_bytes = value_stripped

                validation_error_log = iati.validator.validate_is_xml(value_stripped_bytes)

                if not validation_error_log.contains_errors():
                    self.xml_tree = etree.fromstring(value_stripped_bytes)
                    self._xml_str = value_stripped
                else:
                    if validation_error_log.contains_error_of_type(TypeError):
                        raise TypeError
                    else:
                        raise iati.exceptions.ValidationError(validation_error_log)
            except (AttributeError, TypeError):
                msg = "Datasets can only be ElementTrees or strings containing valid XML, using the xml_tree and xml_str attributes respectively. Actual type: {0}".format(type(value))
                iati.utilities.log_error(msg)
                raise TypeError(msg)

    @property
    def xml_tree(self):
        """ElementTree: A tree representation of the Dataset.

        Raises:
            TypeError: If a value that is being assigned is not an ElementTree.

        Warning:
            Does not fully hide the lxml internal workings.

        Todo:
            Check use of ElementTree in setter.

        """
        return self._xml_tree.getroottree()

    @xml_tree.setter
    def xml_tree(self, value):
        if isinstance(value, etree._Element):  # pylint: disable=W0212
            self._xml_tree = value
            self._xml_str = etree.tostring(value, pretty_print=True)
        else:
            msg = "If setting a Dataset with the xml_property, an ElementTree should be provided, not a {0}.".format(type(value))
            iati.utilities.log_error(msg)
            raise TypeError(msg)

    def _raw_source_at_line(self, line_number):
        """Return the raw value of the XML source at the specified line.

        Args:
            line_number (int): A zero-indexed line number.

        Returns:
            str: The source of the XML at the specified line.

        Raises:
            TypeError: When `line_number` is not an integer.
            ValueError: When `line_number` is negative or more than the number of lines in the file.

        """
        if not isinstance(line_number, int) or isinstance(line_number, bool):
            raise TypeError

        if line_number < 0:
            raise ValueError

        try:
            # this is led with an empty string since the `sourceline` attribute is 1-indexed.
            split_lines = [''] + self.xml_str.split('\n')
            return split_lines[line_number]
        except IndexError:
            raise ValueError

    @property
    def version(self):
        """Return the version of the Standard that this Dataset is specified against.

        Returns:
            str or None: The version of the Standard that this Dataset is specified against. None if the version cannot be detected.

        Todo:
            Consider if this should raise an error if the Dataset is specified at a version of the Standard that does not exist.

        """
        root_tree = self.xml_tree.getroot()
        default_version = '1.01'
        version_iati_root = root_tree.get('version', default_version).strip()

        if version_iati_root.startswith('1'):
            # Version 1 data, so need to check that all child `iati-activity` or `iati-organisation` elements are at the same version
            versions_in_children = list()
            for child_tree in root_tree.getchildren():  # This is expected to return a list of `iati-activity` or `iati-organisation` elements.
                activity_version = child_tree.get('version', default_version).strip()
                versions_in_children.append(activity_version)

            if len(set(versions_in_children)) == 1 and versions_in_children[0] == version_iati_root:
                version = version_iati_root
            else:
                version = None
        else:
            # Not version 1 data, so can return the version specified in `iati-activities/@version`
            version = version_iati_root

        return version

    def source_at_line(self, line_number):
        """Return the value of the XML source at the specified line.

        Args:
            line_number (int): A zero-indexed line number.

        Returns:
            str: The source of the XML at the specified line. Leading and trailing whitespace is trimmed.

        Raises:
            TypeError: When `line_number` is not an integer.
            ValueError: When `line_number` is negative or more than the number of lines in the file.

        Todo:
            Test with minified XML.

        """
        return self._raw_source_at_line(line_number).strip()

    def source_around_line(self, line_number, surrounding_lines=1):
        """Return the value of the XML source at the specified line, plus the specified amount of surrounding context.

        Args:
            line_number (int): A zero-indexed line number.
            surrounding_lines (int): The number of lines of context to provide either side of the specified line number. Default 1.

        Returns:
            str: The source of the XML at the specified line, plus the specified number of lines of surrounding context.
            Should there be fewer lines of XML than are asked for, the entire Dataset will be returned.

        Raises:
            TypeError: When `line_number` is not an integer.
            TypeError: When `surrounding_lines` is not an integer.
            ValueError: When `line_number` is negative or more than the number of lines in the file.
            ValueError: When `surrounding_lines` is negative.

        Todo:
            Test with minified XML.

        """
        if not isinstance(surrounding_lines, int) or isinstance(surrounding_lines, bool):
            raise TypeError

        if surrounding_lines < 0:
            raise ValueError

        lines_arr = []
        lower_line_number = max(line_number - surrounding_lines, 1)
        upper_line_number = min(line_number + surrounding_lines + 1, len(self.xml_str.split('\n')) + 1)

        for line_num in range(lower_line_number, upper_line_number):
            lines_arr.append(self._raw_source_at_line(line_num))

        return '\n'.join(lines_arr)



"""A module containing tests for the library implementation of accessing resources."""
from lxml import etree
import pytest
import six
import iati.constants
import iati.resources
import iati.validator


class TestResources(object):
    """A container for tests relating to resources in general."""

    def test_resource_filename(self):
        """Check that resource file names are found correctly.

        Todo:
            Implement better assertions.

        """
        path = iati.resources.PATH_SCHEMAS
        filename = iati.resources.resource_filename(path)

        assert len(filename) > len(path)
        assert filename.endswith(path)


class TestResourceFolders(object):
    """A container for tests relating to resource folders."""

    @pytest.mark.parametrize('version, expected_version_foldername', [
        ('2.02', '202'),
        ('2.01', '201'),
        ('1.05', '105'),
        ('1.04', '104')
    ])
    def test_get_folder_name_for_version(self, version, expected_version_foldername):
        """Check that expected components are present within folder paths."""
        path = iati.resources.get_folder_name_for_version(version)
        assert expected_version_foldername == path

    @pytest.mark.parametrize('version', [
        '1.00',
        '1',
        1,
        1.01,  # A version must be specified as a string
        'string'
    ])
    def test_get_folder_name_for_version_invalid_version(self, version):
        """Check that an invalid version of the Standard raises a ValueError exception."""
        with pytest.raises(ValueError):
            iati.resources.get_folder_name_for_version(version)

    @pytest.mark.parametrize('path_component', [
        'resources',
        'standard'
    ])
    def test_get_folder_path_for_version(self, standard_version_optional, path_component):
        """Check that expected components are present within folder paths."""
        path = iati.resources.get_folder_path_for_version(*standard_version_optional)
        assert path_component in path

    def test_get_test_data_paths_in_folder(self):
        """Check that test data is being found in specified subfolders.

        Todo:
            Deal with multiple versions.

        """
        paths = iati.resources.get_test_data_paths_in_folder('ssot-activity-xml-fail')

        assert len(paths) == 237


class TestResourceLibraryData(object):
    """A container for tests relating to pyIATI resources."""

    @pytest.mark.parametrize('file_name', [
        'name',
        'Name.xml',
    ])
    def test_get_lib_data_path(self, file_name):
        """Check that library data can be located."""
        path = iati.resources.get_lib_data_path(file_name)

        assert iati.resources.BASE_PATH_LIB_DATA != ''
        assert iati.resources.BASE_PATH_LIB_DATA in path
        assert file_name == path[-len(file_name):]


class TestResourceCodelists(object):
    """A container for tests relating to Codelist resources."""

    def test_codelist_flow_type(self, standard_version_optional):
        """Check that the FlowType codelist is loaded as a string and contains content."""
        path = iati.resources.get_codelist_path('FlowType', *standard_version_optional)

        content = iati.resources.load_as_string(path)

        assert len(content) > 3200
        assert iati.validator.is_xml(content)

    def test_find_codelist_paths(self, codelist_lengths_by_version):
        """Check that all codelist paths are being found."""
        paths = iati.resources.get_all_codelist_paths(codelist_lengths_by_version[0])

        assert len(paths) == codelist_lengths_by_version[1]
        for path in paths:
            assert path[-4:] == iati.resources.FILE_CODELIST_EXTENSION
            assert iati.resources.PATH_CODELISTS in path

    @pytest.mark.parametrize('codelist', [
        'Name',
        'Name.xml',
    ])
    def test_get_codelist_path_name(self, standard_version_optional, codelist):
        """Check that a codelist path is found from just a name."""
        path = iati.resources.get_codelist_path(codelist, *standard_version_optional)

        assert path[-4:] == iati.resources.FILE_CODELIST_EXTENSION
        assert path.count(iati.resources.FILE_CODELIST_EXTENSION) == 1
        assert iati.resources.PATH_CODELISTS in path

    def test_get_codelist_mapping_path(self, standard_version_optional):
        """Check that the Codelist Mapping File path points to a valid XML file."""
        path = iati.resources.get_codelist_mapping_path(*standard_version_optional)

        content = iati.resources.load_as_string(path)

        assert len(content) > 5000
        assert iati.validator.is_xml(content)


class TestResourceSchemas(object):
    """A container for tests relating to Schema resources."""

    def test_get_all_activity_schema_paths(self, standard_version_optional):
        """Check that all activity schema paths are found.

        Todo:
            Handle all paths to schemas being found correctly.

        """
        activity_paths = iati.resources.get_all_activity_schema_paths(*standard_version_optional)

        assert len(activity_paths) == 1

    def test_get_all_organisation_schema_paths(self, standard_version_optional):
        """Check that all organisation schema paths are found.

        Todo:
            Handle all paths to schemas being found correctly.

        """
        organisation_paths = iati.resources.get_all_organisation_schema_paths(*standard_version_optional)

        assert len(organisation_paths) == 1

    def test_find_schema_paths(self, standard_version_optional):
        """Check that both the activity and organisation schema paths are being found.

        Todo:
            Handle all paths to schemas being found correctly.

        """
        paths = iati.resources.get_all_schema_paths(*standard_version_optional)

        assert len(paths) == 2

    @pytest.mark.parametrize('get_schema_path_function', [
        iati.resources.get_all_schema_paths,
        iati.resources.get_all_activity_schema_paths,
        iati.resources.get_all_organisation_schema_paths
    ])
    def test_find_schema_paths_file_extension(self, standard_version_optional, get_schema_path_function):
        """Check that the correct file extension is present within file paths returned by get_all_*schema_paths functions."""
        paths = get_schema_path_function(*standard_version_optional)

        for path in paths:
            assert path[-4:] == iati.resources.FILE_SCHEMA_EXTENSION

    def test_schema_activity_string(self):
        """Check that the Activity schema file contains content."""
        path = iati.resources.get_schema_path('iati-activities-schema')

        content = iati.resources.load_as_string(path)

        assert len(content) > 130000

    def test_schema_activity_tree(self):
        """Check that the Activity schema loads into an XML Tree.

        This additionally involves checking that imported schemas also work.

        """
        path = iati.resources.get_schema_path('iati-activities-schema')
        schema = iati.resources.load_as_tree(path)

        assert isinstance(schema, etree._ElementTree)  # pylint: disable=protected-access


class TestResourceLoading(object):
    """A container for tests relating to loading resources."""

    def test_load_as_bytes(self):
        """Test that resources.load_as_bytes returns a bytes object with the expected content."""
        path_test_data = iati.resources.get_test_data_path('invalid')

        result = iati.resources.load_as_bytes(path_test_data)

        assert isinstance(result, bytes)
        assert result == 'This is a string that is not valid XML\n'.encode()

    def test_load_as_dataset(self):
        """Test that resources.load_as_dataset returns a Dataset object with the expected content."""
        path_test_data = iati.resources.get_test_data_path('valid')

        result = iati.resources.load_as_dataset(path_test_data)

        assert isinstance(result, iati.Dataset)
        assert '<?xml version="1.0"?>\n\n<iati-activities version="2.02">' in result.xml_str

    def test_load_as_dataset_invalid(self):
        """Test that resources.load_as_dataset raises an error when the provided path does not lead to a file containing valid XML."""
        path_test_data = iati.resources.get_test_data_path('invalid')

        with pytest.raises(ValueError):
            _ = iati.resources.load_as_dataset(path_test_data)

    def test_load_as_string(self):
        """Test that `resources.load_as_string()` returns a string (python3) or unicode (python2) object with the expected content."""
        path_test_data = iati.resources.get_test_data_path('invalid')

        result = iati.resources.load_as_string(path_test_data)

        assert isinstance(result, six.string_types)
        assert result == 'This is a string that is not valid XML\n'

    @pytest.mark.parametrize("load_method", [iati.resources.load_as_bytes, iati.resources.load_as_dataset, iati.resources.load_as_string])
    def test_load_as_x_non_existing_file(self, load_method):
        """Test that `resources.load_as_bytes()` returns a bytes object with the expected content."""
        path_test_data = iati.resources.get_test_data_path('this-file-does-not-exist')

        # python 2/3 compatibility - FileNotFoundError introduced at Python 3
        try:
            FileNotFoundError
        except NameError:
            FileNotFoundError = IOError  # pylint: disable=redefined-builtin,invalid-name

        with pytest.raises(FileNotFoundError):
            _ = load_method(path_test_data)

    @pytest.mark.parametrize("file_to_load", [
        'dataset-encoding/valid-windows-1252.xml'
    ])
    def test_load_as_string_restricted_charset(self, file_to_load):
        """Test that Datasets can be loaded from files encoded with a limited charset."""
        path = iati.resources.get_test_data_path(file_to_load)

        data_str = iati.resources.load_as_string(path)
        dataset = iati.Dataset(data_str)
        str_of_interest = dataset.xml_tree.xpath('//reporting-org/narrative/text()')[0]

        # the character of interest is in windows-1252, but is different from ASCII
        # python2/3 compatibility: encode string as UTF-8
        assert str_of_interest.encode('utf-8') == b'\xc5\xb8'

    @pytest.mark.parametrize("file_to_load", [
        'dataset-encoding/valid-UTF-8.xml',
        'dataset-encoding/valid-UTF-16LE.xml',
        'dataset-encoding/valid-UTF-16BE.xml',
        'dataset-encoding/valid-UTF-16.xml',
        'dataset-encoding/valid-UTF-32.xml'
    ])
    def test_load_as_string_unicode(self, file_to_load):
        """Test that Datasets can be loaded from files encoded with various unicode encodings."""
        path = iati.resources.get_test_data_path(file_to_load)

        data_str = iati.resources.load_as_string(path)
        dataset = iati.Dataset(data_str)
        str_of_interest = dataset.xml_tree.xpath('//reporting-org/narrative/text()')[0]

        # the tested characters are all in the code range 004000-00FFFF
        # this means that they are 3-bit at UTF-8, 2 bit as UTF-16 and 4-bit as UTF-32 in 8-bit environments
        # https://en.wikipedia.org/wiki/Comparison_of_Unicode_encodings#Eight-bit_environments
        # python2/3 compatibility: encode string as UTF-8
        assert str_of_interest.encode('utf-8') == b'\xe4\xb6\x8c\xe4\xb9\xa8\xe4\xbc\xb6\xe4\xbe\x97\xe5\x80\x97\xe5\x82\x88\xe5\x84\x88\xe5\x89\x89\xe5\x94\x99\xe8\xac\x9c\xe8\xb0\x8b\xee\x82\xb1\xee\x82\xb2\xee\x82\xb3\xee\x82\xb5\xee\x82\xb8\xee\x82\xba\xee\x82\xbb\xee\x82\xbc\xee\x82\xbd\xee\x82\xbe\xee\x83\x8e\xee\x83\x8f\xee\x84\xa8\xee\x84\xa9\xe4\xb6\x8c\xe4\xb9\xa8\xe4\xbc\xb6\xe4\xbe\x97\xe5\x80\x97\xe5\x82\x88\xe5\x84\x88\xe5\x89\x89\xe5\x94\x99\xe8\xac\x9c\xe8\xb0\x8b\xee\x82\xb1\xee\x82\xb2\xee\x82\xb3\xee\x82\xb5\xee\x82\xb8\xee\x82\xba\xee\x82\xbb\xee\x82\xbc\xee\x82\xbd\xee\x82\xbe\xee\x83\x8e\xee\x83\x8f\xee\x84\xa8\xee\x84\xa9\xe4\xb6\x8c\xe4\xb9\xa8\xe4\xbc\xb6\xe4\xbe\x97\xe5\x80\x97\xe5\x82\x88\xe5\x84\x88\xe5\x89\x89\xe5\x94\x99\xe8\xac\x9c\xe8\xb0\x8b\xee\x82\xb1\xee\x82\xb2\xee\x82\xb3\xee\x82\xb5\xee\x82\xb8\xee\x82\xba\xee\x82\xbb\xee\x82\xbc\xee\x82\xbd\xee\x82\xbe\xee\x83\x8e\xee\x83\x8f\xee\x84\xa8\xee\x84\xa9'  # noqa: ignore=E501  # pylint: disable=line-too-long

    @pytest.mark.parametrize("file_to_load", [
        'dataset-encoding/valid-undetectable-encoding.xml'
    ])
    def test_load_as_string_undetectable_encoding(self, file_to_load):
        """Test that when an encoding cannot be detected, the correct error is raised.

        The file with an undetectable encoding is a UTF-16LE file without a BOM.

        """
        path = iati.resources.get_test_data_path(file_to_load)

        with pytest.raises(ValueError):
            _ = iati.resources.load_as_string(path)



from setuptools import setup, find_packages

setup(
    name = 'pyIATI',
    version = '0.2.0',
    description = 'Python library representing the IATI Schemas, Codelists and Rulesets',
    author = 'IATI Technical Team and other authors',
    author_email = 'code@iatistandard.org',
    url='http://iatistandard.org/',
    packages = find_packages(),
    include_package_data = True,
    install_requires = [
        # detecting character encoding of files
        'chardet==3.0.4',
        # JSON schema parsing validation
        'jsonschema==2.6.0',
        # XML handling library
        'lxml==4.1.0',
        # YAML parsing for validation error codes
        'PyYAML==3.12',
        # python2/python3 compatibility library
        'six==1.11.0'
        ],
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: XML'
        ],
)
